#Copyright 2021 Google LLC

#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
############################################################################################
# -*- coding: utf-8 -*-
################################################################################
#     deep_auto_viml - build and test multiple Tensorflow 2.0 models and pipelines
#     Python v3.6+ tensorflow v2.4.1+
#     Created by Ram Seshadri
#     Licensed under Apache License v2
################################################################################
# data pipelines and feature engg here

# pre-defined TF2 Keras models and your own models here

# Utils
from .classify_features import classify_features_using_pandas
from .classify_features import check_model_options, fast_classify_features
# Utils
from deep_autoviml.utilities.utilities import print_one_row_from_tf_dataset, print_one_row_from_tf_label
from deep_autoviml.utilities.utilities import My_LabelEncoder, print_one_image_from_dataset
from deep_autoviml.utilities.utilities import print_one_text_from_dataset
from deep_autoviml.utilities.utilities import find_columns_with_infinity, drop_rows_with_infinity
############################################################################################
import pandas as pd
import numpy as np
pd.set_option('display.max_columns',500)
from sklearn.model_selection import train_test_split
import matplotlib.pyplot as plt
import tempfile
import pdb
import copy
import warnings
warnings.filterwarnings(action='ignore')
import functools
# Make numpy values easier to read.
np.set_printoptions(precision=3, suppress=True)
# TensorFlow ≥2.4 is required
import tensorflow as tf
np.random.seed(42)
tf.random.set_seed(42)
from tensorflow.keras import layers
from tensorflow import keras
############################################################################################
#### probably the most handy function of all!
def left_subtract(l1,l2):
    lst = []
    for i in l1:
        if i not in l2:
            lst.append(i)
    return lst
import re
def find_words_in_list(words, in_list):
    result = []
    for each_word in words:
        for in_src in in_list:
            if re.findall(each_word, in_src):
                result.append(in_src)
    return list(set(result))

##############################################################################################
def find_problem_type(train, target, model_options={}, verbose=0) :
    """
    ############################################################################
    #####   Now find the problem type of this train dataset using its target variable
    ############################################################################
    """
    target = copy.deepcopy(target)
    ### this determines the number of categories to name integers as classification ##
    ### if a variable has more than this limit, it will not be treated like a cat variable #
    cat_limit = check_model_options(model_options, "variable_cat_limit", 30)
    float_limit = 15 ### this limits the number of float variable categories for it to become cat var
    model_label = 'Single_Label'
    model_class = 'Classification'
    if isinstance(target, str):
        if target == '':
            model_class ='Clustering'
            model_label = 'Single_Label'
            return model_class, model_label,  target
        targ = copy.deepcopy(target)
        target = [target]
    elif isinstance(target, list):
        if len(target) == 1:
            targ = target[0]
        else:
            targ = target[0]
            model_label = 'Multi_Label'
    else:
        print('target is Not detected. Default chosen is %s, %s' %(model_class, model_label))
    ####  This is where you detect what kind of problem it is #################

    if  train[targ].dtype in ['int64', 'int32','int16']:
        if len(train[targ].unique()) <= 2:
            model_class = 'Classification'
        elif len(train[targ].unique()) > 2 and len(train[targ].unique()) <= cat_limit:
            model_class = 'Multi_Classification'
        else:
            model_class = 'Regression'
    elif  train[targ].dtype in ['float']:
        if len(train[targ].unique()) <= 2:
            model_class = 'Classification'
        elif len(train[targ].unique()) > 2 and len(train[targ].unique()) <= float_limit:
            model_class = 'Multi_Classification'
        else:
            model_class = 'Regression'
    else:
        if len(train[targ].unique()) <= 2:
            model_class = 'Classification'
        else:
            model_class = 'Multi_Classification'
    ########### print this for the start of next step ###########
    print('    Model type is %s and %s problem' %(model_class,model_label))
    return model_class, model_label,  target
######################################################################################

def transform_train_target(train_target, target, modeltype, model_label, cat_vocab_dict):
    train_target = copy.deepcopy(train_target)
    cat_vocab_dict = copy.deepcopy(cat_vocab_dict)
    ### Just have to change the target from string to Numeric in entire dataframe! ###
    
    if modeltype != 'Regression':
        if model_label == 'Multi_Label':
            target_copy = copy.deepcopy(target)
            print('Train target shape = %s' %(train_target.shape,))
            #### This is for multi-label problems #####
            cat_vocab_dict['target_le'] = []
            for each_target in target_copy:
                cat_vocab_dict[each_target+'_original_classes'] = np.unique(train_target[target])
                target_le = My_LabelEncoder()
                print('Transforming %s target labels...' %each_target)
                print('    Original target labels data type is %s ' %train_target[each_target].dtype)
                train_values = target_le.fit_transform(train_target[each_target])
                cat_vocab_dict[each_target+'_transformed_classes'] = np.unique(train_values)
                train_target[each_target] = train_values
                cat_vocab_dict['target_le'].append(target_le)
                print('%s transformed as follows: %s' %(each_target, target_le.transformer))
                print('    transformed target labels data type to numeric or ordered from 0')
        else:
            #### This is for Single Label problems ####
            cat_vocab_dict['original_classes'] = np.unique(train_target[target])
            target_le = My_LabelEncoder()
            print('Transforming %s target labels...' %target)
            print('    Original labels dtype is %s ' %train_target[target].dtype)
            train_values = target_le.fit_transform(train_target[target])
            cat_vocab_dict['transformed_classes'] = np.unique(train_values)
            train_target[target] = train_values
            cat_vocab_dict['target_le'] = target_le
            print('%s transformed as follows: %s' %(target, target_le.transformer))
            print('    transformed target labels data type to numeric or ordered from 0')
    else:
        target_le = ""
        cat_vocab_dict['target_le'] = target_le
        print('No Target transformation needed since target dtype is numeric')
    train_target = train_target[target]
    return train_target, cat_vocab_dict

def split_combined_ds_into_two(x, usecols, preds):
    """
    This is useful for splitting a single dataset which has both features and labels into two.
    usecols is basically target column in the form of a list: [target]
    preds is basically predictor columns in the form of a list: a list of predictors
    """
    labels = {k: x[k] for k in x if k in usecols}
    features = {k: x[k] for k in x if k in preds}
    return (features, labels)
######################################################################################################
import pathlib
import os
import random
def load_train_data_file(train_datafile, target, keras_options, model_options, verbose=0):
    """
    This handy function loads a file from a local or remote machine provided the filename and path are given.
    It loads the file(s) into a Tensorflow Dataset using the make_csv_dataset function from Tensorflow 2.0
    """
    train_datafile = copy.deepcopy(train_datafile)
    http_url = False
    if find_words_in_list(['http'], [train_datafile]):
        print('http urls file: will be loaded into pandas and then into tensorflow datasets')
        http_url = True
    try:
        DS_LEN = model_options['DS_LEN']
    except:
        ### Choose a default option in case it is not given
        DS_LEN = 100000
    shuffle_flag = False
    #################################################################################
    try:
        compression = None
        ### see if there is a . in the file name. If it is, then do this process.
        split_str = train_datafile.split(".")[-1]
        if split_str=='csv':
            print("CSV file being loaded into tf.data.Dataset")
            compression_type = None
        elif split_str=='zip' :
            print("Zip file being loaded into tf.data.Dataset")
            compression_type="GZIP" ### don't change this. It is correct.
            compression = "zip" ### don't change this. It is correct.
            print('    Using %s compression_type in make_csv_dataset argument' %compression_type)
        elif split_str=='gz':
            print("Zip file being loaded into tf.data.Dataset")
            compression_type="GZIP"
            compression = "gzip"
            print('    Using %s compression_type in make_csv_dataset argument' %compression_type)
        else:
            compression_type = None
    except:
        #### if . is not there, it means it is a folder and we need to collect all files in that folder
        font_csvs =  sorted(str(p) for p in pathlib.Path(train_datafile).glob("*.csv"))
        print('Printing the first 5 files in the %s folder:\n%s' %(train_datafile,font_csv[:5]))
        train_datafile_list = pathlib.Path(train_datafile).glob("*.csv")
        print('    collecting files matching this file pattern in directory: %s' %train_datafile_list)
        try:
            list_files = []
            filetype = train_datafile.split(".")[-1]
            list_files = [x for x in os.listdir(inpath) if x.endswith(filetype)]
            if list_files == []:
                print('No csv, tsv or Excel files found in the given directory')
                return
            else:
                print('%d files found in directory matching pattern: %s' %(len(list_files), train_datafile))
            ### now you must use this file_pattern in make_csv_dataset argument
            train_datafile = list_files[0]
        except:
            print('not able to collect files matching given pattern = %s' %train_datafile)
            return
    #################################################################################
    model_options['compression'] = compression
    #### About 25% of the data or 10,000 rows which ever is higher is loaded #######
    if http_url:
        maxrows = 100000 ### set it very high so that all rows are read into dataframe ###
    else:
        maxrows = min(100000, int(0.25*DS_LEN))
        print('Max rows loaded to classify features = %s' %maxrows)
    ### first load a small sample of the dataframe and the entire target if it needs transform
    try:
        modeltype = model_options["modeltype"]
    except:
        modeltype, model_label, usecols = find_problem_type(train_small, target, model_options, verbose)
        model_options['modeltype'] = modeltype
    if isinstance(target, str):
        targets = [target]
    else:
        targets = copy.deepcopy(target)
    ######  This is where you select a small sample of a file to do classification of variables ####
    if compression_type:
        ### this reads the entire file and loads it into a dataset if it is a zip file  ######
        train_small = pd.read_csv(train_datafile, sep=sep, nrows=maxrows, compression=compression,
                                header=header, encoding=csv_encoding)
        train_small, data_batches, var_df, cat_vocab_dict, keras_options, model_options = load_train_data_frame(
                        train_small, target, keras_options, model_options, verbose)
        #####  This might be useful for users to know whether to use feature-crosses or not ###
        return train_small, data_batches, var_df, cat_vocab_dict, keras_options, model_options
    else:
        ### It reads only a small dataframe if it is a regular CSV file #######
        train_small = select_rows_from_file_or_frame(train_datafile, model_options, targets, maxrows)
    ##### Now detect modeltype if it is not given ###############
    print('     small sample dataset from train loaded. Shape = %s' %(train_small.shape,))
    #### All column names in Tensorflow should have no spaces ! So you must convert them here!
    sel_preds = ["_".join(x.split(" ")) for x in list(train_small) ]
    header = model_options['header']
    if header is None:
        sel_preds = ["col_"+str(x) for x in range(train_small.shape[1])]
    else:
        sel_preds = ["_".join(x.split("(")) for x in sel_preds ]
        sel_preds = ["_".join(x.split(")")) for x in sel_preds ]
        sel_preds = ["_".join(x.split("/")) for x in sel_preds ]
        sel_preds = ["_".join(x.split("\\")) for x in sel_preds ]
        sel_preds = ["_".join(x.split("?")) for x in sel_preds ]
        sel_preds = [x.lower() for x in sel_preds ]

    if isinstance(target, str):
        target = "_".join(target.split(" "))
        target = "_".join(target.split("("))
        target = "_".join(target.split(")"))
        target = "_".join(target.split("/"))
        target = "_".join(target.split("\\"))
        target = "_".join(target.split("?"))
        target = target.lower()
        model_label = 'Single_Label'
    else:
        target = ["_".join(x.split(" ")) for x in target ]
        target = ["_".join(x.split("(")) for x in target ]
        target = ["_".join(x.split(")")) for x in target ]
        target = ["_".join(x.split("/")) for x in target ]
        target = ["_".join(x.split("\\")) for x in target ]
        target = ["_".join(x.split("?")) for x in target ]
        target = [x.lower() for x in target ]
        model_label = 'Multi_Label'

    train_small.columns = sel_preds

    print('Alert! Modified column names to satisfy rules for column names in Tensorflow...')

    ### modeltype and usecols are very important to know before doing any processing #####
    #### usecols is a very handy tool to handle a target which can be single label or multi-label!
    if modeltype == '':
        ### usecols is basically target in a list format. Very handy to know when target is a list.
        modeltype, _, usecols = find_problem_type(train_small, target, model_options, verbose)
    else:
        ### if modeltype is given, then do not find the model type using this function
        _,  _, usecols = find_problem_type(train_small, target, model_options, verbose)

    
    ##########  Find small details about the data to help create the right model ###
    label_encode_flag = model_options["label_encode_flag"]
    if isinstance(label_encode_flag, str):
        if modeltype == 'Classification' or modeltype == 'Multi_Classification':
            if isinstance(target, str):
                #### This is for Single-Label problems ########
                if train_small[target].dtype == 'object' or str(train_small[target].dtype).lower() == 'category':
                    label_encode_flag = True
                elif 0 not in np.unique(train_small[target]):
                    label_encode_flag = False 
                    if verbose:
                        print('    label encoding must be done since there is no zero class!')
                target_vocab = train_small[target].unique()
                num_classes = len(target_vocab)
            elif isinstance(target, list):
                #### This is for Multi-Label problems ########
                num_classes = []
                for each_target in target:
                    if train_small[each_target].dtype == 'object' or str(train_small[target[0]].dtype).lower() == 'category':
                        label_encode_flag = True
                    elif 0 not in np.unique(train_small[each_target]):
                        label_encode_flag = False
                        if verbose:
                            print('    label encoding must be done since there is no zero class!')
                    target_vocab = train_small[each_target].unique().tolist()
                    num_classes.append(len(target_vocab))
        else:
            num_classes = 1
            target_vocab = []
            label_encode_flag =  False
    else:
        if isinstance(target, str):
            target_vocab = train_small[target].unique()
            num_classes = len(target_vocab)
        else:
            for each_target in copy_target:
                target_vocab = train_small[target].unique().tolist()
                num_classes_each = len(target_vocab)
                num_classes.append(int(num_classes_each))
                
    #### This is where we set the model_options for num_classes and num_labels #########
    model_options['num_classes'] = num_classes
    
    #############   Sample Data classifying features into variaous types ##################
    print('Loaded a small data sample of size = %s into pandas dataframe to analyze...' %(train_small.shape,))
    ### classify variables using the small dataframe ##
    print('    Classifying variables using data sample in pandas...')
    train_small, var_df1, cat_vocab_dict = classify_features_using_pandas(train_small, target, model_options, verbose=verbose)
    
    ##########    Just transfer all the values from var_df to cat_vocab_dict  ##################################
    for each_key in var_df1:
        cat_vocab_dict[each_key] = var_df1[each_key]
    ############################################################################################################

    model_options['modeltype'] = modeltype
    model_options['model_label'] = model_label
    cat_vocab_dict['modeltype'] = modeltype
    cat_vocab_dict['target_variables'] = usecols
    cat_vocab_dict['num_classes'] = num_classes
    cat_vocab_dict["target_transformed"] = label_encode_flag

    # Construct a lookup table to map string chars to indexes,

    # using the vocab loaded above:
    if label_encode_flag:
        #### Sometimes, using tf.int64 works. Hence this is needed.
        table = tf.lookup.StaticHashTable(
            tf.lookup.KeyValueTensorInitializer(
                keys=target_vocab, values=tf.constant(list(range(len(target_vocab))),
                                               dtype=tf.int64)),
            default_value=int(len(target_vocab)+1))

    #### Set column defaults while reading dataset from CSV files - that way, missing values avoided!
    ### The following are valid CSV dtypes for missing values: float32, float64, int32, int64, or string
    ### fill all missing values in categorical variables with "None"
    ### Similarly. fill all missing values in float variables with -99
    if train_small.isnull().sum().sum() > 0:
        print('    %d missing values in dataset: filling them with default values...' %(
                                train_small.isnull().sum().sum()))
    string_cols = train_small.select_dtypes(include='object').columns.tolist() + train_small.select_dtypes(
                                        include='category').columns.tolist()
    integer_cols =  train_small.select_dtypes(include='integer').columns.tolist()
    float_cols = train_small.select_dtypes(include='float').columns.tolist()
    bool_cols = train_small.select_dtypes(include='bool').columns.tolist()
    ### Bool_columns become string after you set their defaults since missing is default ##
    column_defaults = [-99.0 if x in float_cols else -99 if x in integer_cols else "missing"
                                for x in list(train_small)]
    ### So we need to put back bool columns as boolean right after we load them into data_batches
    ####### Make sure you don't move this next stage. It should be after column defaults! ###
    if label_encode_flag:
        trans_output, cat_vocab_dict = transform_train_target(train_small, target, modeltype,
                                    model_label, cat_vocab_dict)
        train_small[target] = trans_output.values

    #### CAUTION: (num_epochs=None) will automatically repeat the data forever! Be Careful with it!
    ### setting num_epochs to 1 is always good practice since it ensures that your dataset is readable later
    ###  If you set num_epochs to None it will throw your dataset batches into infinite loop. Be careful!
    ####  Also the dataset will display the batch size as 4 (or whatever) if you set num_epochs as None.
    ####  However, if you set num_epochs=1, then you will see dataset shape as None!
    ####  Also num_epochs=1 need to do repeat() on the dataset to loop it forever.
    num_epochs = 1

    ########### find the number of labels in data ####
    if isinstance(target, str):
        num_labels = 1
    elif isinstance(target, list):
        if len(target) == 1:
            num_labels = 1
        else:
            num_labels = len(target)
    cat_vocab_dict['num_labels'] = num_labels
    model_options['num_labels'] = num_labels

    ###  Initially set this batch_size low so that you can do better model training with small batches ###
    #### It is a good idea to shuffle and batch it with small batch size like 4 immediately ###
    if http_url:
        ### Once a file is in gzip format, you have to load it into pandas and then find file size and batch
        cat_vocab_dict["DS_LEN"] = train_small.shape[0]
        model_options['DS_LEN'] = train_small.shape[0]
        DS_LEN = train_small.shape[0]
    try:
        keras_options["batchsize"] = batch_size
        if isinstance(keras_options["batchsize"], str):
            batch_size = find_batch_size(DS_LEN)
        cat_vocab_dict['batch_size'] = batch_size
    except:
        batch_size = find_batch_size(DS_LEN)
        keras_options["batchsize"] = batch_size
        cat_vocab_dict['batch_size'] = batch_size
    ######  Do this for selecting what columns to load into TF.Data  #######
    #### This means it is not a test dataset - hence it has target columns - load it too!
    if isinstance(target, str):
        if target == '':
            target_name = None
        else:
            target_name = copy.deepcopy(target)
            preds = [x for x in list(train_small) if x not in [target]]
    elif isinstance(target, list):
        #### then it is a multi-label problem
        target_name = None
        preds = left_subtract(list(train_small), target)
    else:
        print('Error: Target %s type not understood' %type(target))
        return

    ############################################################################################
    ###########  C H E C K   F O R   BOOL and I N F I N I T E   V A L U E S  H E R E ###########
    ############################################################################################
    cols_with_infinity = find_columns_with_infinity(train_small)
    @tf.function
    def convert_boolean_to_string(features, target):
        """
        This is how you convert all your boolean features into float variables.
        The reason you have to do this is because tf.keras does not know how to handle boolean types.
        It takes as input an ordered dict named features and returns the same in features format.
        """
        for feature_name in features:
            if feature_name in bool_cols:
                # Cast boolean feature values to string.
                #features[feature_name] = tf.cast(features[feature_name], tf.dtypes.float32)
                features[feature_name] = tf.dtypes.cast(features[feature_name], tf.string)
        return (features, target)

    ################    T F  D A T A   D A T A S E T   L O A D I N G     H E R E ################
    ############    Create a Tensorflow Dataset using the make_csv function #####################
    if http_url:
        print('Since input is http URL file we load it into pandas and then tf.data.Dataset...')
        ### Now load the URL file loaded into pandas into a tf.data.dataset  #############
        if isinstance(target, str):
            if target != '':
                labels = train_small.pop(target)
                data_batches = tf.data.Dataset.from_tensor_slices((dict(train_small), labels))
            else:
                print('target variable is blank - please fix input and try again')
                return
        elif isinstance(target, list):
            ##### For multi-label problems, you need to use dict of labels as well ###
            labels = train_small.pop(target)
            data_batches = tf.data.Dataset.from_tensor_slices((dict(train_small), dict(labels)))
        else:
            data_batches = tf.data.Dataset.from_tensor_slices(dict(train_small))
        ### batch it if you are creating it from a dataframe
        data_batches = data_batches.batch(batch_size, drop_remainder=True)
    else:
        print('Loading your input file(s) data directly into tf.data.Dataset...')
        data_batches = tf.data.experimental.make_csv_dataset(train_datafile,
                                       batch_size=batch_size,
                                       column_names=sel_preds,
                                       label_name=target_name,
                                       num_epochs = num_epochs,
                                       column_defaults=column_defaults,
                                       compression_type=compression_type,
                                       shuffle=shuffle_flag,
                                       num_parallel_reads=tf.data.experimental.AUTOTUNE)
        ############### Additional post-processing checkes needed - do it here  #######
        #### here is where we need to put back boolean columns that were strings back to boolean
        if bool_cols:
            data_batches = data_batches.map(convert_boolean_to_string)
        ### Remove this not after testing the function below ###
        if cols_with_infinity:
            data_batches = data_batches.map(drop_non_finite_rows)
            print('    ALERT! Dropping non-finite values in %d columns: %s ' %(
                                            len(cols_with_infinity), cols_with_infinity))

        ########   P E R F O R M   L A B E L   E N C O D I N G   H E R E ############
        if label_encode_flag:
            print('    target label encoding now...')
            data_batches = data_batches.map(lambda x, y: to_ids(x, y, table))
            print('    target label encoding completed.')
    print('    train data loaded successfully.')
    drop_cols = var_df1['cols_delete']
    preds = [x for x in list(train_small) if x not in usecols+drop_cols]
    print('\nNumber of predictors to be used = %s in predict step: keras preprocessing...' %len(preds))
    cat_vocab_dict['columns_deleted'] = drop_cols
    if len(drop_cols) > 0: ### drop cols that have been identified for deletion ###
        print('Dropping %s columns marked for deletion...' %drop_cols)
        train_small.drop(drop_cols,axis=1,inplace=True)
    model_options['train_data_is_file'] = True
    return train_small, data_batches, var_df1, cat_vocab_dict, keras_options, model_options
############################################################################################
def drop_non_finite_rows(features, targets):
  cols = []
  for key, col in features.items():
    cols.append(col)
  # stack the columns to build a matrix
  cols = tf.stack(cols, axis=-1)
  # The good rows are the ones where all the elements are finite
  good = tf.reduce_all(tf.math.is_finite(cols), axis=-1)

  # Apply the boolean mask to each column and return it as a dict.
  result = {}
  for name, value in features.items():
    result[name] = tf.boolean_mask(value,good)
  return result, targets
############################################################################################
def to_ids(features, labels, table):
    if labels.dtype==np.int32: labels = tf.cast(labels, tf.int64)
    #labels = tf.cast(labels, tf.int64) ## this should not have been used ##
    labels = table.lookup(labels)
    return (features, labels)
#############################################################################################
def lenopenreadlines(filename):
    with open(filename) as f:
        return len(f.readlines())
#########################################################################################
def closest(lst, K):
    """
    Find a number in list lst that is closest to the value K.
    """
    return lst[min(range(len(lst)), key = lambda i: abs(lst[i]-K))]
##########################################################################################
def find_batch_size(DS_LEN):
    ### Since you cannot deal with a very large dataset in pandas, let's look into how big the file is
    maxrows = 10000
    if DS_LEN < 100:
        batch_ratio = 0.16
    elif DS_LEN >= 100 and DS_LEN < 1000:
        batch_ratio = 0.05
    elif DS_LEN >= 1000 and DS_LEN < 10000:
        batch_ratio = 0.01
    elif DS_LEN >= maxrows and DS_LEN <= 100000:
        batch_ratio = 0.001
    else:
        batch_ratio = 0.0001
    batch_len = int(batch_ratio*DS_LEN)
    #print('    Batch size selected as %d' %batch_len)
    lst = [32, 48, 64, 96, 128, 256]
    batch_len = closest(lst, batch_len)
    return batch_len
#########################################################################################
def fill_missing_values_for_TF2(train_small, var_df):
    """
    ########################################################################################
    ### As of now (TF 2.4.1) we still cannot load pd.dataframe with nulls in string columns!
    ### You must first remove nulls from the objects in dataframe and use that sample
    ### to build a normalizer layer. You can use Mean and SD from that sample.
    ### Using that sample, you can build the layer for complete dataset
    #### in that case the input is a dataframe, you must first remove nulls from it
    ########################################################################################
    ### Filling Strategy (this is not Imputation - mind you)
    ###    1. Fill all missing values in categorical variables with "None"
    ###    2. Similarly, fill all missing values in float variables with -99
    ########################################################################################
    """
    train_small = copy.deepcopy(train_small)
    bools = var_df['bools']
    cols_delete = var_df['cols_delete']
    cat_cols = var_df['categorical_vars'] + var_df['discrete_string_vars'] + bools
    int_bools = var_df['int_bools']
    int_cats = var_df['int_cats']
    ints = var_df['int_vars']
    int_cols = int_cats + ints + int_bools
    float_cols = var_df['continuous_vars']
    nlp_cols = var_df['nlp_vars']
    date_vars = var_df['date_vars']
    lats = var_df['lat_vars']
    lons = var_df['lon_vars']
    ffill_cols = lats + lons + date_vars

    if len(cat_cols) > 0:
        if train_small[cat_cols].isnull().sum().sum() > 0:
            for col in cat_cols:
                colcount = "Missing"
                train_small[col].fillna(colcount, inplace=True)

    if len(nlp_cols) > 0:
        if train_small[nlp_cols].isnull().sum().sum() > 0:
            for col in nlp_cols:
                colcount = "Missing"
                train_small[col].fillna(colcount, inplace=True)

    ints_copy = int_cols + int_cats
    if len(ints_copy) > 0:
        if train_small[ints_copy].isnull().sum().sum() > 0:
            for col in ints_copy:
                colcount = 0
                train_small[col].fillna(colcount,inplace=True)

    if len(float_cols) > 0:
        if train_small[float_cols].isnull().sum().sum() > 0:
            for col in float_cols:
                colcount = 0.0
                train_small[col].fillna(colcount,inplace=True)

    ffill_cols = train_small.columns[train_small.isnull().sum()>0]

    if len(ffill_cols) > 0:
        ffill_cols_copy = copy.deepcopy(ffill_cols)
        if train_small[ffill_cols].isnull().sum().sum() > 0:
            for col in ffill_cols:
                train_small[col].fillna(method='ffill', inplace=True)
        #### Sometimes forward-fill doesn't do it. You need to try back-fill too!
        if train_small[ffill_cols].isnull().sum().sum() > 0:
            for col in ffill_cols_copy:
                train_small[col].fillna(method='bfill', inplace=True)
    return train_small
########################################################################################
def load_train_data_frame(train_dataframe, target, keras_options, model_options, verbose=0):
    """
    ### CAUTION: TF2.4 Still cannot load a DataFrame with Nulls in string or categoricals!
    ############################################################################
    #### TF 2.4 still cannot load tensor_slices into ds if an object or string column
    ####   that has nulls in it! So we need to find other ways to load tensor_slices by
    ####   first filling dataframe with pandas fillna() function!
    #############################################################################
    """
    train_dataframe = copy.deepcopy(train_dataframe)
    DS_LEN = model_options['DS_LEN']
    print('Max rows loaded to classify features = %s' %DS_LEN)
    print('     small sample dataset from train loaded. Shape = %s' %(train_dataframe.shape,))
    #### do this for dataframes ##################
    maxrows = 100000
    try:
        batch_size = keras_options["batchsize"]
        if isinstance(keras_options["batchsize"], str):
            batch_size = find_batch_size(DS_LEN)
    except:
        #### If it is not given find it here ####
        batch_size = find_batch_size(DS_LEN)
    #########  Modify or Convert column names to fit tensorflow rules of no space in names!
    sel_preds = ["_".join(x.split(" ")) for x in list(train_dataframe) ]
    #### This can also be a problem with other special characters ###
    sel_preds = ["_".join(x.split("(")) for x in sel_preds ]
    sel_preds = ["_".join(x.split(")")) for x in sel_preds ]
    sel_preds = ["_".join(x.split("/")) for x in sel_preds ]
    sel_preds = ["_".join(x.split("\\")) for x in sel_preds ]
    sel_preds = ["_".join(x.split("?")) for x in sel_preds ]
    sel_preds = [x.lower() for x in sel_preds ]

    if isinstance(target, str):
        target = "_".join(target.split(" "))
        target = "_".join(target.split("("))
        target = "_".join(target.split(")"))
        target = "_".join(target.split("/"))
        target = "_".join(target.split("\\"))
        target = "_".join(target.split("?"))
        target = target.lower()
        model_label = 'Single_Label'
    else:
        target = ["_".join(x.split(" ")) for x in target ]
        target = ["_".join(x.split("(")) for x in target ]
        target = ["_".join(x.split(")")) for x in target ]
        target = ["_".join(x.split("/")) for x in target ]
        target = ["_".join(x.split("\\")) for x in target ]
        target = ["_".join(x.split("?")) for x in target ]
        target = [x.lower() for x in target ]
        model_label = 'Multi_Label'

    train_dataframe.columns = sel_preds

    print('Alert! Modified column names to satisfy rules for column names in Tensorflow...')


    #### if target is changed you must send that modified target back to other processes ######
    ### usecols is basically target in a list format. Very handy to know when target is a list.
    
    try:
        modeltype = model_options["modeltype"]
        if model_options["modeltype"] == '':
            ### usecols is basically target in a list format. Very handy to know when target is a list.
            modeltype, model_label, usecols = find_problem_type(train_dataframe, target, model_options, verbose)
            model_options["modeltype"] = modeltype
        else:
            if isinstance(target, str):
                usecols = [target]
            else:
                usecols = copy.deepcopy(target)
    except:
        ### if modeltype is given, then do not find the model type using this function
        modeltype,  model_label, usecols = find_problem_type(train_dataframe, target, model_options, verbose)

    if isinstance(target, str):
        targets = [target]
    else:
        targets = copy.deepcopy(target)

    #####  This is a simple function to load a small sample of data to do analysis ############
    train_small = select_rows_from_file_or_frame(train_dataframe, model_options, targets, maxrows)

    ###   Cat_Vocab_Dict contains all info about vocabulary in each variable and their size
    print('    Classifying variables using data sample in pandas...')   
    train_small, var_df, cat_vocab_dict = classify_features_using_pandas(train_small, target, model_options, verbose=verbose)

    ##########    Just transfer all the values from var_df to cat_vocab_dict  ##################################
    for each_key in var_df:
        cat_vocab_dict[each_key] = var_df[each_key]
    ############################################################################################################
    model_options['modeltype'] = modeltype
    model_options['model_label'] = model_label
    cat_vocab_dict['target_variables'] = usecols
    cat_vocab_dict['modeltype'] = modeltype
    model_options['batch_size'] = batch_size
    ##########  Find small details about the data to help create the right model ###
    target_transformed = model_options["label_encode_flag"]
    if isinstance(target_transformed, str):
        if modeltype != 'Regression':
            if isinstance(target, str):
                #### This is for Single Label Problems ######
                if train_small[target].dtype == 'object' or str(train_small[target].dtype).lower() == 'category':
                    target_transformed = True
                    target_vocab = train_small[target].unique()
                    num_classes = len(target_vocab)
                else:
                    if 0 not in np.unique(train_small[target]):
                        target_transformed = True ### label encoding must be done since no zero class!
                        target_vocab = train_small[target].unique()
                    num_classes = len(train_small[target].value_counts())
            elif isinstance(target, list):
                #### This is for Multi-Label Problems #######
                copy_target = copy.deepcopy(target)
                num_classes = []
                for each_target in copy_target:
                    if train_small[target[0]].dtype == 'object' or str(train_small[target[0]].dtype).lower() == 'category':
                        target_transformed = True
                        target_vocab = train_small[target].unique().tolist()
                        num_classes_each = len(target_vocab)
                    else:
                        if 0 not in np.unique(train_small[target[0]]):
                            target_transformed = True ### label encoding must be done since no zero class!
                            target_vocab = train_small[target[0]].unique()
                        num_classes_each = train_small[target].apply(np.unique).apply(len).max()
                    num_classes.append(int(num_classes_each))
        else:
            num_classes = 1
            target_vocab = []
            target_transformed = False
    else:
        if isinstance(target, str):
            target_vocab = train_small[target].unique()
            num_classes = len(target_vocab)
        else:
            for each_target in copy_target:
                target_vocab = train_small[target].unique().tolist()
                num_classes_each = len(target_vocab)
                num_classes.append(int(num_classes_each))

    ########### find the number of labels in data ####
    if isinstance(target, str):
        num_labels = 1
    elif isinstance(target, list):
        if len(target) == 1:
            num_labels = 1
        else:
            num_labels = len(target)
    #### This is where we set the model_options for num_classes and num_labels #########
    model_options['num_labels'] = num_labels
    model_options['num_classes'] = num_classes
    cat_vocab_dict['num_labels'] = num_labels
    cat_vocab_dict['num_classes'] = num_classes
    cat_vocab_dict["target_transformed"] = target_transformed
    
    #### once the dataframe has been classified, you can again change train_small to original dataframe ##
    train_small = copy.deepcopy(train_dataframe)

    ####   fill missing values using this function ##############
    train_small = fill_missing_values_for_TF2(train_small, cat_vocab_dict)

    ##### Do the deletion of cols after filling with missing values since otherwise fill errors!
    drop_cols = var_df['cols_delete']
    cat_vocab_dict['columns_deleted'] = drop_cols
    if len(drop_cols) > 0: ### drop cols that have been identified for deletion ###
        print('    Dropping %s columns marked for deletion...' %drop_cols)
        train_small.drop(drop_cols,axis=1,inplace=True)

    ######### Now load the train Dataframe into a tf.data.dataset  #############
    if target_transformed:
        ####################### T R A N S F O R M I N G   T A R G E T ########################
        train_small[target], cat_vocab_dict = transform_train_target(train_small, target, modeltype,
                                                model_label, cat_vocab_dict)

    if isinstance(target, str):
        #### For single label do this: labels can be without names since there is only one label
        if target != '':
            labels = train_small[target]
            features = train_small.drop(target, axis=1)
            ds = tf.data.Dataset.from_tensor_slices((dict(features), labels))
        else:
            print('target variable is blank - please fix input and try again')
            return
    elif isinstance(target, list):
        #### For multi label do this: labels must be dict and hence with names since there are many targets
            labels = train_small[target]
            features = train_small.drop(target, axis=1)
            ds = tf.data.Dataset.from_tensor_slices((dict(features), dict(labels)))
    else:
        ds = tf.data.Dataset.from_tensor_slices(dict(train_small))
    ######   Now save some defaults in cat_vocab_dict ##########################
    try:
        keras_options["batchsize"] = batch_size
        cat_vocab_dict['batch_size'] = batch_size
    except:
        batch_size = find_batch_size(DS_LEN)
        keras_options["batchsize"] = batch_size
        cat_vocab_dict['batch_size'] = batch_size
    
    ##########################################################################
    #### C H E C K  F O R  I N F I N I T E   V A L U E S   H E R E ##########
    ##########################################################################
    cols_with_infinity = find_columns_with_infinity(train_small)
    if cols_with_infinity:
        train_small = drop_rows_with_infinity(train_small, cols_with_infinity, fill_value=True)
    model_options['train_data_is_file'] = False
    return train_small, ds, var_df, cat_vocab_dict, keras_options, model_options
###############################################################################################
def load_image_data(image_directory, project_name, keras_options, model_options,
                        verbose=0):
    """
    Handy function that collects a sequence of image files  into a tf.data generator.

    Your images input folder or directory must be like the following. If not, you will get error.
        main_directory/
        ...class_a/
        ......image_1.png
        ......image_2.png
        ...class_b/
        ......image_3.png
        ......image_4.png

    Inputs:
    -----------
    image_directory: This is the folder that contains image files organized by class_label.
    project_name: This is where the model will be stored once it is trained.
    keras_options: a data dictionary that contains keras model options you can send.
    model_options: a data dictionary that saves all the characteristics of your model

    Outputs:
    -----------
    train_ds: a train dataset in tf.data.Dataset
    valid_ds: a validation dataset in tf.data.Dataset format
    cat_vocab_dict: a data dictionary that saves all the characteristics of your data
    model_options: a data dictionary that saves all the characteristics of your model
    """
    cat_vocab_dict = dict()
    cat_vocab_dict['target_variables'] =  "target"
    cat_vocab_dict['project_name'] = project_name
    if 'image_height' in model_options.keys():
        print('    Image height given as %d' %model_options['image_height'])
    else:
        print("    No image height given. Returning. Provide image height and width...")
        return
    if 'image_width' in model_options.keys():
        print('    Image width given as %d' %model_options['image_width'])
    else:
        print("    No image width given. Returning. Provide image height and width...")
        return
    if 'image_channels' in model_options.keys():
        print('    Image channels given as %d' %model_options['image_channels'])
    else:
        print("    No image_channels given. Returning. Provide image height and width...")
        return
    try:
        image_train_folder = os.path.join(image_directory,"train")
        if not os.path.exists(image_train_folder):
            print("Image use case. No train folder exists under given directory. Returning...")
            return
    except:
        print('Error: Not able to find any train or test image folder in the given folder %s' %image_train_folder)
        print("""   You must put images under folders named train,
                    validation (optional) and test folders under given %s folder.
                    Otherwise deep_autoviml won't work. """ %image_directory)
    image_train_split = False
    image_train_folder = os.path.join(image_directory,"train")
    if not os.path.exists(image_train_folder):
        print("No train folder found under given image directory %s. Returning..." %image_directory)
        image_train_folder = os.path.join(image_directory,"validation")
    image_valid_folder = os.path.join(image_directory,"validation")
    if not os.path.exists(image_valid_folder):
        print("No validation folder found under given image directory %s. Returning..." %image_directory)
        image_train_split = True
    img_height = model_options['image_height']
    img_width = model_options['image_width']
    img_channels = model_options['image_channels']
    #### make this a small number - default batch_size ###
    batch_size = check_model_options(model_options, "batch_size", 64)
    model_options["batch_size"] = batch_size
    full_ds = tf.keras.preprocessing.image_dataset_from_directory(image_train_folder,
                                  seed=111,
                                  image_size=(img_height, img_width),
                                  batch_size=batch_size)
    if image_train_split:
        ############## Split train into train and validation datasets here ###############
        classes = full_ds.class_names
        recover = lambda x,y: y
        print('\nSplitting train into two: train and validation data')
        valid_ds = full_ds.enumerate().filter(is_valid).map(recover)
        train_ds = full_ds.enumerate().filter(is_train).map(recover)
    else:
        train_ds = full_ds
        valid_ds = tf.keras.preprocessing.image_dataset_from_directory(image_valid_folder,
          seed=111,
          image_size=(img_height, img_width),
          batch_size=batch_size)
        classes = train_ds.class_names
    ####  Successfully loaded train and validation data sets ################
    cat_vocab_dict["image_classes"] = classes
    cat_vocab_dict["target_transformed"] = True
    cat_vocab_dict['modeltype'] =  'Classification'
    MLB = My_LabelEncoder()
    ins = copy.deepcopy(classes)
    outs = np.arange(len(classes))
    MLB.transformer = dict(zip(ins,outs))
    MLB.inverse_transformer = dict(zip(outs,ins))
    cat_vocab_dict['target_le'] = MLB
    print('Number of image classes = %d and they are: %s' %(len(classes), classes))
    if len(classes) <= 2:
        model_options["num_predicts"] = 1
    else:
        model_options["num_predicts"] = len(classes)
    print_one_image_from_dataset(train_ds, classes)
    AUTOTUNE = tf.data.AUTOTUNE
    train_ds = train_ds.cache().prefetch(buffer_size=AUTOTUNE)
    valid_ds = valid_ds.cache().prefetch(buffer_size=AUTOTUNE)
    cat_vocab_dict["image_height"] = img_height
    cat_vocab_dict["image_width"] = img_width
    cat_vocab_dict["batch_size"] = batch_size
    cat_vocab_dict["image_channels"] = img_channels
    return train_ds, valid_ds, cat_vocab_dict, model_options
########################################################################################
from collections import defaultdict
from collections import Counter

def load_train_data(train_data_or_file, target, project_name, keras_options, model_options,
                  keras_model_type, verbose=0):
    """
    Handy function that loads a file or a sequence of files (*.csv) into a tf.data.Dataset
    You can also load a pandas dataframe instead of a file if you wanted to. It accepts both!
    It will automatically figure out whether input is a file or file(s) or a pandas dataframe.

    Inputs: train_data_or_file, target
    -------------------------------------------------------------------------------
    train_data_or_file: this can be a name of file to load or can be a pandas dataframe to load into tf.data
                  either option will work. This function will detect that automatically and load them.
    target: target name as a string or a list

    Outputs: train_small, model_options, ds, var_df, cat_vocab_dict, keras_options
    -------------------------------------------------------------------------------
    train_small: a sample of data into a pandas dataframe
    model_options: a dictionary describing the data
    ds: a tf.data.Dataset containing a symbolic link to the data at rest in your train_data_or_file
    var_df: a dictionary classifying features in data to multiple types such as numeric, category, etc.
    cat_vocab_dict: a dictionary containing artifacts from the data that will be used during inference
    keras_options: a dictionary containing keras defaults for the model that will be built using this data
    """
    shuffle_flag = False
    cat_vocab_dict = defaultdict(list)
    train_data_or_file = copy.deepcopy(train_data_or_file)
    maxrows = 10000 ### the number of maximum rows read by pandas to sample data ##
    ### Since you cannot deal with a very large dataset in pandas, let's look into how big the file is
    try:
        if isinstance(train_data_or_file, str):
            DS_LEN = lenopenreadlines(train_data_or_file)
        else:
            DS_LEN = train_data_or_file.shape[0]
    except:
        if find_words_in_list(['http'], [train_data_or_file.lower()]):
            print('http url file: cannot find size of dataset. Setting default...')
        DS_LEN = maxrows #### set to an artificial low number ###
    keras_options["data_size"] = DS_LEN
    model_options["DS_LEN"] = DS_LEN
    ##########    LOADING EITHER FILE OR DATAFRAME INTO TF DATASET HERE  ##################
    if isinstance(train_data_or_file, str):
        #### do this for files only ##################
        train_small, train_ds, var_df, cat_vocab_dict, keras_options, model_options = load_train_data_file(train_data_or_file, target,
                                                                keras_options, model_options, verbose)
    else:
        train_small, train_ds, var_df, cat_vocab_dict, keras_options, model_options = load_train_data_frame(train_data_or_file, target,
                                                                keras_options, model_options, verbose)

    ### This is where we do all kinds of feature engineering - this needs to be in predict ####
    cat_vocab_dict['bools_converted'] = False
    if isinstance(train_data_or_file, str):
        ### if train_data is a file, boolean vars have to be converted to strings ###
        BOOLS = []
        cat_vocab_dict['bools_converted'] = True
        cat_vocab_dict['categorical_vars'] += cat_vocab_dict['bools']
    else:
        ### if train is a dataframe, you can leave bools as it is ###
        BOOLS = cat_vocab_dict['bools']
    #################################################################################
    ##### F E A T U R E    E N G I N E E R I N G   H E R E              #############
    #################################################################################
    def process_boolean(features, target):
        """
        This is how you convert all your boolean features into float variables.
        The reason you have to do this is because tf.keras does not know how to handle boolean types.
        It takes as input an ordered dict named features and returns the same in features format.
        """
        for feature_name in features:
            if feature_name in BOOLS:
                # Cast boolean feature values to int32 only if the train_data is a dataframe ##
                #features[feature_name] = tf.cast(features[feature_name], tf.dtypes.float32)
                features[feature_name] = tf.dtypes.cast(features[feature_name], tf.int32)
        return (features, target)
    #################################################################################
    train_ds = train_ds.map(process_boolean)
    #################################################################################
    ################## process boolean target if needed #############################
    #################################################################################
    #@tf.autograph.experimental.do_not_convert
    @tf.function
    def process_target(features, target):
        target = tf.cast(target, tf.dtypes.float32)
        return (features, target)
    if bool in train_small[cat_vocab_dict['target_variables']].dtypes.values:
        train_ds = train_ds.map(process_target)        
    print('Boolean columns successfully processed')
    #################################################################################
    if keras_model_type.lower() in ['nlp', 'text']:
        NLP_VARS = cat_vocab_dict['predictors_in_train']
    else:
        NLP_VARS = cat_vocab_dict['nlp_vars']
    ################################################################
    @tf.function
    def process_NLP_features(features):
        """
        This is how you combine all your string NLP features into a single new feature.
        Then you can perform embedding on this combined feature.
        It takes as input an ordered dict named features and returns the same features format.
        """
        return tf.strings.reduce_join([features[i] for i in NLP_VARS],axis=0,
                keepdims=False, separator=' ', name="combined")
    ################################################################
    NLP_COLUMN = "combined_nlp_text"
    ################################################################
    @tf.function
    def combine_nlp_text(features):
        ##use x to derive additional columns u want. Set the shape as well
        y = {}
        y.update(features)
        y[NLP_COLUMN] = tf.strings.reduce_join([features[i] for i in NLP_VARS],axis=0,
                keepdims=False, separator=' ')
        return y
    ######################################################################################
    ### You have to load only the NLP or text variables into dataset. 
    ###    otherwise, it will fail during predict. Yo still need to create input for them.
    ###  In mixed_NLP models, you drop original NLP vars and combine them into one NLP var.
    ######################################################################################
    
    if NLP_VARS and keras_model_type.lower() in ['nlp','text', 'mixed_nlp', 'combined_nlp']:
        if keras_model_type.lower() in ['nlp', 'text']:
            train_ds = train_ds.map(lambda x, y: (process_NLP_features(x), y))
            #train_ds = train_ds.unbatch().batch(batch_size)
            print('    processed NLP or text vars: %s successfully' %NLP_VARS)
        elif keras_model_type.lower() in ['combined_nlp']:
            train_ds = train_ds.map(lambda x, y: (combine_nlp_text(x), y))
            print('    combined NLP or text vars: %s into a single feature successfully' %NLP_VARS)
        else:
            ### Mixed NLP is to keep NLP vars separate so they can be processed individually ##
            print('    keeping NLP vars separate')
    else:
        print('     No special text preprocessing done for NLP vars.')
    ############################################################################
    ### You must batch it if you are creating it from a dataframe
    batch_size = cat_vocab_dict['batch_size']
    if not isinstance(train_data_or_file, str):
        train_ds = train_ds.batch(batch_size, drop_remainder=True)
    #### if Target is modified in the above processes such as removing spaces, etc. you must re-init here
    usecols = cat_vocab_dict['target_variables']
    cat_vocab_dict['DS_LEN'] = DS_LEN
    if verbose >= 1 and train_small.shape[1] <= 30:
        print_one_row_from_tf_dataset(train_ds)
    ####  Set Class Weights for Imbalanced Data Sets here ##########
    modeltype = model_options["modeltype"]
    #### You need to do this transform only for files. Otherwise, it is done already for dataframes.
    if len(usecols) == 1:
        target = usecols[0]
        ### This is a single label problem ########
        y_train = train_small[target]
        if modeltype != 'Regression' and not cat_vocab_dict['target_transformed']:
            cat_vocab_dict["original_classes"] = np.unique(train_small[target])
    else:
        ### This is a Multi-label label problem ########
        target = usecols[0]
        y_train = train_small[usecols[0]]
        target_copy = copy.deepcopy(usecols)
        if modeltype != 'Regression' and not cat_vocab_dict['target_transformed']:
            for each_t in target_copy:
                cat_vocab_dict[each_t+"_original_classes"] = np.unique(train_small[each_t])
    ####  CREATE  CLASS_WEIGHTS HERE #################
    if modeltype != 'Regression':
        find_rare_class(y_train, verbose=1)
        if 'class_weight' in keras_options.keys() and not model_options['model_label']=='Multi_Label':
            # Class weights are only applicable to single labels in Keras right now
            class_weights = get_class_distribution(y_train)
            keras_options['class_weight'] = class_weights
            print('    Class weights calculated: %s' %class_weights)
        else:
            keras_options['class_weight'] = {}
    else:
        keras_options['class_weight'] = {}
        print('    No class weights specified. Continuing...')
    return train_small, model_options, train_ds, var_df, cat_vocab_dict, keras_options
##########################################################################################################
from collections import OrderedDict
def find_rare_class(classes, verbose=0):
    ######### Print the % count of each class in a Target variable  #####
    """
    Works on Multi Class too. Prints class percentages count of target variable.
    It returns the name of the Rare class (the one with the minimum class member count).
    This can also be helpful in using it as pos_label in Binary and Multi Class problems.
    """
    counts = OrderedDict(Counter(classes))
    total = sum(counts.values())
    if verbose >= 1:
        print('       Class  -> Counts -> Percent')
        sorted_keys = sorted(counts.keys())
        for cls in sorted_keys:
            print("%12s: % 7d  ->  % 5.1f%%" % (cls, counts[cls], counts[cls]/total*100))
    if type(pd.Series(counts).idxmin())==str:
        return pd.Series(counts).idxmin()
    else:
        return int(pd.Series(counts).idxmin())
###############################################################################
from sklearn.utils.class_weight import compute_class_weight
import copy
from collections import Counter
def get_class_distribution(y_input):
    y_input = copy.deepcopy(y_input)
    classes = np.unique(y_input)
    xp = Counter(y_input)
    class_weights = compute_class_weight('balanced', classes=np.unique(y_input), y=y_input)
    if len(class_weights[(class_weights> 10)]) > 0:
        class_weights = (class_weights/10)
    else:
        class_weights = (class_weights)
    #print('    class_weights = %s' %class_weights)
    class_weights[(class_weights<1)]=1
    class_rows = class_weights*[xp[x] for x in classes]
    class_rows = class_rows.astype(int)
    class_weighted_rows = dict(zip(classes,class_weights))
    #print('    class_weighted_rows = %s' %class_weighted_rows)
    return class_weighted_rows
########################################################################
### Split raw_train_set into train and valid data sets first
### This is a better way to split a dataset into train and test ####
### It does not assume a pre-defined size for the data set.
def is_valid(x, y):
    return x % 5 == 0
def is_test(x, y):
    return x % 2 == 0
def is_train(x, y):
    return not is_test(x, y)
##################################################################################
def load_text_data(text_directory, project_name, keras_options, model_options,
                        verbose=0):
    """
    Handy function that collects a sequence of text files  into a tf.data generator.
    Your text input folder or directory must be like the following. If not, you will get error.
        main_directory/
        ...class_a/
        ......a_text_1.txt
        ......a_text_2.txt
        ...class_b/
        ......b_text_1.txt
        ......b_text_2.txt

    Inputs:
    -----------
    text_directory: This is the folder that contains .txt files organized by class_label.
    project_name: This is where the model will be stored once it is trained.
    keras_options: a data dictionary that contains keras model options you can send.
    model_options: a data dictionary that saves all the characteristics of your model

    Outputs:
    -----------
    train_ds: a train dataset in tf.data.Dataset
    valid_ds: a validation dataset in tf.data.Dataset format
    cat_vocab_dict: a data dictionary that saves all the characteristics of your data
    model_options: a data dictionary that saves all the characteristics of your model
    """
    cat_vocab_dict = dict()
    cat_vocab_dict['target_variables'] =  "target"
    cat_vocab_dict['project_name'] = project_name
    try:
        text_train_folder = os.path.join(text_directory,"train")
        if not os.path.exists(text_train_folder):
            print("text use case. No train folder exists under given directory. Returning...")
            return
    except:
        print('Error: Not able to find any train or test folder in the given folder %s' %text_train_folder)
        print("""   You must put texts under folders named train,
                    validation (optional) and test folders under given %s folder.
                    Otherwise deep_autoviml won't work. """ %text_directory)
    text_train_split = False
    text_train_folder = os.path.join(text_directory,"train")
    if not os.path.exists(text_train_folder):
        print("No train folder found under given text directory %s. Returning..." %text_directory)
        text_train_folder = os.path.join(text_directory,"validation")
    text_valid_folder = os.path.join(text_directory,"validation")
    if not os.path.exists(text_valid_folder):
        print("No validation folder found under given text directory %s. Returning..." %text_directory)
        text_train_split = True
    #### make this a small number - default batch_size ###
    batch_size = check_model_options(model_options, "batch_size", 64)
    model_options["batch_size"] = batch_size
    full_ds = tf.keras.preprocessing.text_dataset_from_directory(text_train_folder,
                                  seed=111,
                                  batch_size=batch_size)
    if text_train_split:
        ############## Split train into train and validation datasets here ###############
        classes = full_ds.class_names
        recover = lambda x,y: y
        print('\nSplitting train into two: train and validation data')
        valid_ds = full_ds.enumerate().filter(is_valid).map(recover)
        train_ds = full_ds.enumerate().filter(is_train).map(recover)
    else:
        train_ds = full_ds
        valid_ds = tf.keras.preprocessing.text_dataset_from_directory(text_valid_folder,
          seed=111,
          batch_size=batch_size)
        classes = train_ds.class_names
    ####  Successfully loaded train and validation data sets ################
    cat_vocab_dict["text_classes"] = classes
    cat_vocab_dict["target_transformed"] = True
    cat_vocab_dict['modeltype'] =  'Classification'
    MLB = My_LabelEncoder()
    ins = copy.deepcopy(classes)
    outs = np.arange(len(classes))
    MLB.transformer = dict(zip(ins,outs))
    MLB.inverse_transformer = dict(zip(outs,ins))
    cat_vocab_dict['target_le'] = MLB
    print('Number of text classes = %d and they are: %s' %(len(classes), classes))
    print_one_text_from_dataset(train_ds, classes)
    AUTOTUNE = tf.data.AUTOTUNE
    train_ds = train_ds.cache().prefetch(buffer_size=AUTOTUNE)
    valid_ds = valid_ds.cache().prefetch(buffer_size=AUTOTUNE)
    model_options["num_classes"] = len(classes)
    cat_vocab_dict["batch_size"] = batch_size
    return train_ds, valid_ds, cat_vocab_dict, model_options
###################################################################################
def select_rows_from_file_or_frame(train_datafile, model_options, targets, nrows_limit):
    train_datafile = copy.deepcopy(train_datafile)
    #### Set some defaults from model options which is required ##
    DS_LEN = model_options["DS_LEN"]
    sep = model_options["sep"]
    header = model_options["header"]
    csv_encoding = model_options["csv_encoding"]
    modeltype = model_options['modeltype']
    compression = model_options['compression']
    ####### we randomly sample a small dataset to classify features #####################
    test_size = min(0.9, (1 - (nrows_limit/DS_LEN))) ### make sure there is a small train size    
    if test_size <= 0:
        test_size = 0.9
    if DS_LEN > nrows_limit:
        print('    Since number of rows > %s, loading a random sample of %d rows into pandas for EDA' %(nrows_limit, DS_LEN))
    ######  If it is a file you need to load it into a dataframe, it not leave it as is ###
    if isinstance(train_datafile, str):
        ###### load a small sample of data into a pandas dataframe ##
        if DS_LEN >= 1e5:
            train_small = pd.read_csv(train_datafile, nrows=nrows_limit, sep=sep, header=header,
                            encoding=csv_encoding, compression=compression)
        else:
            train_small = pd.read_csv(train_datafile, sep=sep, header=header,
                            encoding=csv_encoding, compression=compression)
    else:
        train_small = copy.deepcopy(train_datafile)
    ####### If it is a classification problem, you need to stratify and select sample ###
    
    if modeltype != 'Regression':
        copy_targets = copy.deepcopy(targets)
        for each_target in copy_targets:
            ### You need to remove rows that have very class samples - that is a problem while splitting train_small
            list_of_few_classes = train_small[each_target].value_counts()[train_small[each_target].value_counts()<=10].index.tolist()
            train_small = train_small.loc[~(train_small[each_target].isin(list_of_few_classes))]
        train_small, _ = train_test_split(train_small, test_size=test_size, stratify=train_small[targets])
    else:
        ### For Regression problems: load a small sample of data into a pandas dataframe ##
        if DS_LEN <= nrows_limit:
            train_small = train_small.sample(n=DS_LEN, random_state=99)
        else:
            train_small = train_small.sample(n=nrows_limit, random_state=99)
    return train_small
######################################################################################


############################################################################################
#Copyright 2021 Google LLC

#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
############################################################################################
import tensorflow as tf
import numpy as np
import matplotlib.pyplot as plt
import logging

logging.getLogger('tensorflow').setLevel(logging.ERROR)

from tensorflow.keras.callbacks import Callback
#########################################################################################################
######   One Cycle is a Super-Convergence technique developed by Leslie Smith: https://arxiv.org/abs/1708.07120
######                Super-Convergence: Very Fast Training of Neural Networks Using Large Learning Rates
######   This particular implementation is by Andrich van Wyk • September 02, 2019
######   Used with permission: https://www.avanwyk.com/tensorflow-2-super-convergence-with-the-1cycle-policy/
#########################################################################################################
class CosineAnnealer:
    
    def __init__(self, start, end, steps):
        self.start = start
        self.end = end
        self.steps = steps
        self.n = 0
        
    def step(self):
        self.n += 1
        cos = np.cos(np.pi * (self.n / self.steps)) + 1
        return self.end + (self.start - self.end) / 2. * cos


class OneCycleScheduler(Callback):
    """ 
    #########################################################################################################
    ######   One Cycle is a Super-Convergence technique developed by Leslie Smith: https://arxiv.org/abs/1708.07120
    ######                Super-Convergence: Very Fast Training of Neural Networks Using Large Learning Rates
    ######   This particular implementation is by Andrich van Wyk • September 02, 2019
    ######   Credit: https://www.avanwyk.com/tensorflow-2-super-convergence-with-the-1cycle-policy/
    #########################################################################################################
    Callback that schedules the learning rate on a 1cycle policy as per Leslie Smith's paper(https://arxiv.org/pdf/1803.09820.pdf).
    If the model supports a momentum parameter, it will also be adapted by the schedule.
    The implementation adopts additional improvements as per the fastai library: https://docs.fast.ai/callbacks.one_cycle.html, where
    only two phases are used and the adaptation is done using cosine annealing.
    """

    def __init__(self, lr_max, steps, mom_min=0.85, mom_max=0.95, phase_1_pct=0.3, div_factor=25.):
        super(OneCycleScheduler, self).__init__()
        lr_min = lr_max / div_factor
        final_lr = lr_max / (div_factor * 1e4)
        phase_1_steps = steps * phase_1_pct
        phase_2_steps = steps - phase_1_steps
        
        self.phase_1_steps = phase_1_steps
        self.phase_2_steps = phase_2_steps
        self.phase = 0
        self.step = 0
        
        self.phases = [[CosineAnnealer(lr_min, lr_max, phase_1_steps), CosineAnnealer(mom_max, mom_min, phase_1_steps)], 
                 [CosineAnnealer(lr_max, final_lr, phase_2_steps), CosineAnnealer(mom_min, mom_max, phase_2_steps)]]
        
        self.lrs = []
        self.moms = []

    def on_train_begin(self, logs=None):
        self.phase = 0
        self.step = 0

        self.set_lr(self.lr_schedule().start)
        self.set_momentum(self.mom_schedule().start)
        
    def on_train_batch_begin(self, batch, logs=None):
        self.lrs.append(self.get_lr())
        self.moms.append(self.get_momentum())

    def on_train_batch_end(self, batch, logs=None):
        self.step += 1
        if self.step >= self.phase_1_steps:
            self.phase = 1
        self.set_lr(self.lr_schedule().step())
        self.set_momentum(self.mom_schedule().step())
        
    def get_lr(self):
        try:
            return tf.keras.backend.get_value(self.model.optimizer.lr)
        except AttributeError:
            return None
        
    def get_momentum(self):
        try:
            return tf.keras.backend.get_value(self.model.optimizer.momentum)
        except AttributeError:
            return None
        
    def set_lr(self, lr):
        try:
            if lr < 0:
                lr = 0.1
                self.phase = 0
                self.step = 0

                self.set_lr(self.lr_schedule().start)
                self.set_momentum(self.mom_schedule().start)
                tf.keras.backend.clear_session()

            tf.keras.backend.set_value(self.model.optimizer.lr, lr)
        except AttributeError:
            pass # ignore
        
    def set_momentum(self, mom):
        try:
            tf.keras.backend.set_value(self.model.optimizer.momentum, mom)
        except AttributeError:
            pass # ignore

    def lr_schedule(self):
        return self.phases[self.phase][0]
    
    def mom_schedule(self):
        return self.phases[self.phase][1]
    
    def plot(self):
        ax = plt.subplot(1, 2, 1)
        ax.plot(self.lrs)
        ax.set_title('Learning Rate')
        ax = plt.subplot(1, 2, 2)
        ax.plot(self.moms)
        ax.set_title('Momentum')


#Copyright 2021 Google LLC

#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
############################################################################################
# -*- coding: utf-8 -*-
################################################################################
#     deep_autoviml - build and test multiple Tensorflow 2.0 models and pipelines
#     Python v3.6+ tensorflow v2.4.1+
#     Created by Ram Seshadri
#     Licensed under Apache License v2
################################################################################
import pandas as pd
import numpy as np
np.random.seed(99)
#### The warnings from Sklearn are so annoying that I have to shut it off #######
import warnings
warnings.filterwarnings("ignore")
from sklearn.exceptions import DataConversionWarning
warnings.filterwarnings(action='ignore', category=DataConversionWarning)
def warn(*args, **kwargs):
    pass
warnings.warn = warn
####################################################################################
import re
import pdb
import pprint
from itertools import cycle, combinations
from collections import defaultdict, OrderedDict
import time
import sys
import random
import xlrd
import statsmodels
from io import BytesIO
import base64
from functools import reduce
import copy
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import tempfile
import copy
import warnings
warnings.filterwarnings(action='ignore')
import functools
# Make numpy values easier to read.
np.set_printoptions(precision=3, suppress=True)
############################################################################################
# data pipelines and feature engg here

# pre-defined TF2 Keras models and your own models here

# Utils

############################################################################################
# TensorFlow ≥2.4 is required
import tensorflow as tf
np.random.seed(42)
tf.random.set_seed(42)
from tensorflow.keras import layers
from tensorflow import keras
from tensorflow.keras.layers.experimental.preprocessing import Normalization, StringLookup
from tensorflow.keras.layers.experimental.preprocessing import IntegerLookup, CategoryEncoding
from tensorflow.keras.layers.experimental.preprocessing import TextVectorization

from tensorflow.keras.optimizers import SGD, Adam, RMSprop
from tensorflow.keras import layers
from tensorflow.keras import optimizers
from tensorflow.keras.models import Model, load_model
from tensorflow.keras import callbacks
from tensorflow.keras import backend as K
from tensorflow.keras import utils
from tensorflow.keras.layers import BatchNormalization
from tensorflow.keras.optimizers import SGD
from tensorflow.keras import regularizers

from sklearn.metrics import roc_auc_score, mean_squared_error, mean_absolute_error
from IPython.core.display import Image, display
import pickle

##### Suppress all TF2 and TF1.x warnings ###################
tf2logger = tf.get_logger()
tf2logger.warning('Silencing TF2.x warnings')
tf2logger.root.removeHandler(tf2logger.root.handlers)
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
############################################################################################
from tensorflow.keras.layers import Reshape, MaxPooling1D, MaxPooling2D, AveragePooling2D, AveragePooling1D
from tensorflow.keras import Model, Sequential
from tensorflow.keras.layers import Activation, Dense, Embedding, GlobalAveragePooling1D, GlobalMaxPooling1D, Dropout, Conv1D
from tensorflow.keras.layers.experimental.preprocessing import TextVectorization

#######################################################################################################
def classify_features(dfte, depVar, model_options={}, verbose=0):
    max_cols_analyzed = 30
    dfte = copy.deepcopy(dfte)
    if isinstance(depVar, list):
        orig_preds = [x for x in list(dfte) if x not in depVar]
    else:
        orig_preds = [x for x in list(dfte) if x not in [depVar]]
    #################    CLASSIFY  COLUMNS   HERE    ######################
    preds_copy_copy = copy.deepcopy(orig_preds)
    for key in preds_copy_copy:
        if len(dfte[key].map(type).value_counts()) > 1:
            print('Alert! %s has %d mixed data types: %s ' %(key,len(dfte[key].map(type).value_counts()),
                                                dfte[key].map(type).value_counts().index))
    var_df = classify_columns(dfte[orig_preds], model_options, verbose)
    #####       Classify Columns   ################
    IDcols = var_df['id_vars']
    nlp_vars = var_df['nlp_vars']
    discrete_string_vars = var_df['discrete_string_vars']
    cols_delete = var_df['cols_delete']
    int_vars = var_df['int_vars'] + var_df['num_bool_vars']
    categorical_vars = var_df['cat_vars'] + var_df['factor_vars'] +  var_df['string_bool_vars']
    date_vars = var_df['date_vars']
    continuous_vars = var_df['continuous_vars']
    ####### Now search for latitude and longitude variables ######
    lats, lons, matched_pairs = find_latitude_longitude_columns_in_df(dfte[orig_preds], verbose)
    if len(lats+lons) > 0:
        continuous_vars = left_subtract(continuous_vars, lats+lons)
        categorical_vars = left_subtract(categorical_vars, lats+lons)
        discrete_string_vars = left_subtract(discrete_string_vars, lats+lons)
    ######################################################################
    #cols_delete = list(set(IDcols+cols_delete))  ## leave IDcols in dataset. We need ID's to track rows.
    preds = [x for x in orig_preds if x not in cols_delete]

    var_df['cols_delete'] = cols_delete
    if len(cols_delete) == 0:
        print('        No variables removed since no ID or low-information variables found in data set')
    else:
        print('        %d variable(s) to be removed since they were ID or low-information variables'
                                %len(cols_delete))
        if verbose >= 1:
            print('    List of variables to be removed: %s' %cols_delete)
    #############  Check if there are too many columns to visualize  ################
    ppt = pprint.PrettyPrinter(indent=4)
    if verbose > 1 and len(preds) <= max_cols_analyzed:
        marthas_columns(dfte,verbose)
        print("   Columns to delete:")
        ppt.pprint('   %s' % cols_delete)
        print("   Categorical variables: ")
        ppt.pprint('   %s' % categorical_vars)
        print("   Continuous variables:" )
        ppt.pprint('   %s' % continuous_vars)
        print("   Discrete string variables: " )
        ppt.pprint('   %s' % discrete_string_vars)
        print("   NLP string variables: " )
        ppt.pprint('   %s' % nlp_vars)
        print("   Date and time variables: " )
        ppt.pprint('   %s' % date_vars)
        if len(lats) > 0:
            print("   Latitude variables:" )
            ppt.pprint('   %s' % lats)
        if len(lons) > 0:
            print("   Longitude variables:" )
            ppt.pprint('   %s' % lons)
        if len(matched_pairs) > 0:
            print("   Matched Latitude and Longitude variables:" )
            ppt.pprint('   %s' % matched_pairs)
        print("   ID variables %s ")
        ppt.pprint('   %s' % IDcols)
        print("   Target variable %s ")
        ppt.pprint('   %s' % depVar)
    elif verbose==1 and len(preds) > max_cols_analyzed:
        print('   Total columns > %d, too numerous to list.' %max_cols_analyzed)
    features_dict = dict([('IDcols',IDcols),('cols_delete',cols_delete),('categorical_vars',categorical_vars), (
                        'lat_vars',lats),('lon_vars',lons),('matched_pairs',matched_pairs), ('int_vars',int_vars),
                        ('continuous_vars',continuous_vars),('discrete_string_vars',discrete_string_vars),
                        ('nlp_vars',nlp_vars), ('date_vars',date_vars)])
    return features_dict
#######################################################################################################
def marthas_columns(data,verbose=0):
    """
    This program is named  in honor of my one of students who came up with the idea for it.
    It's a neat way of printing data types and information compared to the boring describe() function in Pandas.
    """
    data = data[:]
    print('Data Set Shape: %d rows, %d cols' % data.shape)
    if data.shape[1] > 30:
        print('Too many columns to print')
    else:
        if verbose==1:
            print('Data Set columns info:')
            for col in data.columns:
                print('* %s: %d nulls, %d unique vals, most common: %s' % (
                        col,
                        data[col].isnull().sum(),
                        data[col].nunique(),
                        data[col].value_counts().head(2).to_dict()
                    ))
            print('--------------------------------------------------------------------')
################################################################################
######### NEW And FAST WAY to CLASSIFY COLUMNS IN A DATA SET #######
################################################################################
from collections import defaultdict
def classify_columns(df_preds, model_options={}, verbose=0):
    """
    This actually does Exploratory data analysis - it means this function performs EDA
    ######################################################################################
    Takes a dataframe containing only predictors to be classified into various types.
    DO NOT SEND IN A TARGET COLUMN since it will try to include that into various columns.
    Returns a data frame containing columns and the class it belongs to such as numeric,
    categorical, date or id column, boolean, nlp, discrete_string and cols to delete...
    ####### Returns a dictionary with 10 kinds of vars like the following: # continuous_vars,int_vars
    # cat_vars,factor_vars, bool_vars,discrete_string_vars,nlp_vars,date_vars,id_vars,cols_delete
    """
    train = copy.deepcopy(df_preds)
    #### If there are 30 chars are more in a discrete_string_var, it is then considered an NLP variable
    ### if a variable has more than this many chars, it will be treated like a NLP variable

    max_nlp_char_size = check_model_options(model_options, "nlp_char_limit", 50)
    ### if a variable has more than this limit, it will not be treated like a cat variable #
    #### Cat_Limit defines the max number of categories a column can have to be called a categorical colum
    cat_limit = check_model_options(model_options, "variable_cat_limit", 30)
    max_cols_to_print = 30
    #### Make this limit low so that float variables below this limit become cat vars ###
    float_limit = 15
    print('############## C L A S S I F Y I N G  V A R I A B L E S  ####################')
    print('Classifying variables in data set...')
    def add(a,b):
        return a+b
    sum_all_cols = defaultdict(list)
    orig_cols_total = train.shape[1]
    #Types of columns
    cols_delete = [col for col in list(train) if (len(train[col].value_counts()) == 1
                                   ) | (train[col].isnull().sum()/len(train) >= 0.90)]
    train = train[left_subtract(list(train),cols_delete)]
    var_df = pd.Series(dict(train.dtypes)).reset_index(drop=False).rename(
                        columns={0:'type_of_column'})
    sum_all_cols['cols_delete'] = cols_delete
    var_df['bool'] = var_df.apply(lambda x: 1 if x['type_of_column'] in ['bool','object']
                        and len(train[x['index']].value_counts()) == 2 else 0, axis=1)
    string_bool_vars = list(var_df[(var_df['bool'] ==1)]['index'])
    sum_all_cols['string_bool_vars'] = string_bool_vars
    var_df['num_bool'] = var_df.apply(lambda x: 1 if x['type_of_column'] in [np.uint8,
                            np.uint16, np.uint32, np.uint64,
                            'int8','int16','int32','int64',
                            'float16','float32','float64'] and len(
                        train[x['index']].value_counts()) == 2 else 0, axis=1)
    num_bool_vars = list(var_df[(var_df['num_bool'] ==1)]['index'])
    sum_all_cols['num_bool_vars'] = num_bool_vars
    ######   This is where we take all Object vars and split them into diff kinds ###
    discrete_or_nlp = var_df.apply(lambda x: 1 if x['type_of_column'] in ['object']  and x[
        'index'] not in string_bool_vars+cols_delete else 0,axis=1)
    ######### This is where we figure out whether a string var is nlp or discrete_string var ###
    var_df['nlp_strings'] = 0
    var_df['discrete_strings'] = 0
    var_df['cat'] = 0
    var_df['id_col'] = 0
    var_df['date_time'] = 0
    discrete_or_nlp_vars = var_df.loc[discrete_or_nlp==1]['index'].values.tolist()
    ###### This is where we detect categorical variables based on category limit #######
    if len(var_df.loc[discrete_or_nlp==1]) != 0:
        for col in discrete_or_nlp_vars:
            #### first fill empty or missing vals since it will blowup ###
            train[col] = train[col].fillna('  ')
            if train[col].map(lambda x: len(x) if type(x)==str else 0).mean(
                ) >= max_nlp_char_size and len(train[col].value_counts()
                        ) >= int(0.9*len(train)) and col not in string_bool_vars:
                try:
                    pd.to_datetime(train[col],infer_datetime_format=True)
                    var_df.loc[var_df['index']==col,'date_time'] = 1
                except:
                    var_df.loc[var_df['index']==col,'nlp_strings'] = 1
            elif len(train[col].value_counts()) > cat_limit and len(train[col].value_counts()
                        ) <= int(0.9*len(train)) and col not in string_bool_vars:
                try:
                    pd.to_datetime(train[col],infer_datetime_format=True)
                    var_df.loc[var_df['index']==col,'date_time'] = 1
                except:
                    var_df.loc[var_df['index']==col,'discrete_strings'] = 1
            elif len(train[col].value_counts()) > cat_limit and len(train[col].value_counts()
                        ) == len(train) and col not in string_bool_vars:
                try:
                    pd.to_datetime(train[col],infer_datetime_format=True)
                    var_df.loc[var_df['index']==col,'date_time'] = 1
                except:
                    var_df.loc[var_df['index']==col,'id_col'] = 1
            else:
                var_df.loc[var_df['index']==col,'cat'] = 1
    nlp_vars = list(var_df[(var_df['nlp_strings'] ==1)]['index'])
    sum_all_cols['nlp_vars'] = nlp_vars
    discrete_string_vars = list(var_df[(var_df['discrete_strings'] ==1) ]['index'])
    sum_all_cols['discrete_string_vars'] = discrete_string_vars
    date_vars = list(var_df[(var_df['date_time'] == 1)]['index'])
    ###### This happens only if a string column happens to be an ID column #######
    #### DO NOT Add this to ID_VARS yet. It will be done later.. Dont change it easily...
    #### Category DTYPE vars are very special = they can be left as is and not disturbed in Python. ###
    var_df['dcat'] = var_df.apply(lambda x: 1 if str(x['type_of_column'])=='category' else 0,
                            axis=1)
    factor_vars = list(var_df[(var_df['dcat'] ==1)]['index'])
    sum_all_cols['factor_vars'] = factor_vars
    ########################################################################
    date_or_id = var_df.apply(lambda x: 1 if x['type_of_column'] in [np.uint8,
                         np.uint16, np.uint32, np.uint64,
                         'int8','int16',
                        'int32','int64']  and x[
        'index'] not in string_bool_vars+num_bool_vars+discrete_string_vars+nlp_vars+date_vars else 0,
                                        axis=1)
    ######### This is where we figure out whether a numeric col is date or id variable ###
    var_df['int'] = 0
    ### if a particular column is date-time type, now set it as a date time variable ##
    var_df['date_time'] = var_df.apply(lambda x: 1 if x['type_of_column'] in ['<M8[ns]','datetime64[ns]']  and x[
        'index'] not in string_bool_vars+num_bool_vars+discrete_string_vars+nlp_vars else 1 if x['date_time']==1 else 0,
                                        axis=1)
    ### this is where we save them as date time variables ###
    if len(var_df.loc[date_or_id==1]) != 0:
        for col in var_df.loc[date_or_id==1]['index'].values.tolist():
            if len(train[col].value_counts()) == len(train):
                if train[col].min() < 1900 or train[col].max() > 2050:
                    var_df.loc[var_df['index']==col,'id_col'] = 1
                else:
                    try:
                        pd.to_datetime(train[col],infer_datetime_format=True)
                        var_df.loc[var_df['index']==col,'date_time'] = 1
                    except:
                        var_df.loc[var_df['index']==col,'id_col'] = 1
            else:
                if train[col].min() < 1900 or train[col].max() > 2050:
                    if col not in num_bool_vars:
                        var_df.loc[var_df['index']==col,'int'] = 1
                else:
                    try:
                        pd.to_datetime(train[col],infer_datetime_format=True)
                        var_df.loc[var_df['index']==col,'date_time'] = 1
                    except:
                        if col not in num_bool_vars:
                            var_df.loc[var_df['index']==col,'int'] = 1
    else:
        pass
    int_vars = list(var_df[(var_df['int'] ==1)]['index'])
    date_vars = list(var_df[(var_df['date_time'] == 1)]['index'])
    id_vars = list(var_df[(var_df['id_col'] == 1)]['index'])
    sum_all_cols['int_vars'] = int_vars
    copy_date_vars = copy.deepcopy(date_vars)
    ###### In Tensorflow there is no need to create age variables from year-dates. Hence removing them!
    for date_var in copy_date_vars:
        if train[date_var].dtype in ['int16','int32','int64']:
            if train[date_var].min() >= 1900 or train[date_var].max() <= 2050:
                ### if it is between these numbers, its probably a year so avoid adding it
                date_items = train[date_var].dropna(axis=0).apply(str).apply(len).values
                if all(date_items[0] == item for item in date_items):
                    if date_items[0] == 4:
                        print('    Changing %s from date-var to int-var' %date_var)
                        int_vars.append(date_var)
                        date_vars.remove(date_var)
                        continue
        else:
            date_items = train[date_var].dropna(axis=0).apply(str).apply(len).values
            #### In some extreme cases, 4 digit date variables are not useful
            if all(date_items[0] == item for item in date_items):
                if date_items[0] == 4:
                    print('    Changing %s from date-var to discrete-string-var' %date_var)
                    discrete_string_vars.append(date_var)
                    date_vars.remove(date_var)
                    continue
        #### This test is to make sure sure date vars are actually date vars
        try:
            pd.to_datetime(train[date_var],infer_datetime_format=True)
        except:
            ##### if not a date var, then just add it to delete it from processing
            cols_delete.append(date_var)
            date_vars.remove(date_var)
    sum_all_cols['date_vars'] = date_vars
    sum_all_cols['id_vars'] = id_vars
    sum_all_cols['cols_delete'] = cols_delete
    ## This is an EXTREMELY complicated logic for cat vars. Don't change it unless you test it many times!
    var_df['numeric'] = 0
    float_or_cat = var_df.apply(lambda x: 1 if x['type_of_column'] in ['float16',
                            'float32','float64'] else 0,
                                        axis=1)
    if len(var_df.loc[float_or_cat == 1]) > 0:
        for col in var_df.loc[float_or_cat == 1]['index'].values.tolist():
            if len(train[col].value_counts()) > 2 and len(train[col].value_counts()
                ) <= float_limit and len(train[col].value_counts()) <= len(train):
                var_df.loc[var_df['index']==col,'cat'] = 1
            else:
                if col not in num_bool_vars:
                    var_df.loc[var_df['index']==col,'numeric'] = 1
    cat_vars = list(var_df[(var_df['cat'] ==1)]['index'])
    continuous_vars = list(var_df[(var_df['numeric'] ==1)]['index'])
    ########  V E R Y    I M P O R T A N T   ###################################################
    ##### There are a couple of extra tests you need to do to remove abberations in cat_vars ###
    cat_vars_copy = copy.deepcopy(cat_vars)
    for cat in cat_vars_copy:
        if df_preds[cat].dtype==float:
            continuous_vars.append(cat)
            cat_vars.remove(cat)
            var_df.loc[var_df['index']==cat,'cat'] = 0
            var_df.loc[var_df['index']==cat,'numeric'] = 1
        elif len(df_preds[cat].value_counts()) == df_preds.shape[0]:
            id_vars.append(cat)
            cat_vars.remove(cat)
            var_df.loc[var_df['index']==cat,'cat'] = 0
            var_df.loc[var_df['index']==cat,'id_col'] = 1
    sum_all_cols['cat_vars'] = cat_vars
    sum_all_cols['continuous_vars'] = continuous_vars
    sum_all_cols['id_vars'] = id_vars
    cols_delete = find_remove_duplicates(cols_delete+id_vars)
    sum_all_cols['cols_delete'] = cols_delete
    ###### This is where you consoldate the numbers ###########
    var_dict_sum = dict(zip(var_df.values[:,0], var_df.values[:,2:].sum(1)))
    for col, sumval in var_dict_sum.items():
        if sumval == 0:
            print('%s of type=%s is not classified' %(col,train[col].dtype))
        elif sumval > 1:
            print('%s of type=%s is classified into more then one type' %(col,train[col].dtype))
        else:
            pass
    ###############  This is where you print all the types of variables ##############
    ####### Returns 8 vars in the following order: continuous_vars,int_vars,cat_vars,
    ###  string_bool_vars,discrete_string_vars,nlp_vars,date_or_id_vars,cols_delete
    cat_vars_copy = copy.deepcopy(cat_vars)
    for each_cat in cat_vars_copy:
        if len(train[each_cat].value_counts()) > cat_limit:
            discrete_string_vars.append(each_cat)
            cat_vars.remove(each_cat)
    sum_all_cols['cat_vars'] = cat_vars
    sum_all_cols['discrete_string_vars'] = discrete_string_vars
    #########  The variables can now be printed ##############

    if verbose == 1:
        print("    Number of Numeric Columns = ", len(continuous_vars))
        print("    Number of Integer-Categorical Columns = ", len(int_vars))
        print("    Number of String-Categorical Columns = ", len(cat_vars))
        print("    Number of Factor-Categorical Columns = ", len(factor_vars))
        print("    Number of String-Boolean Columns = ", len(string_bool_vars))
        print("    Number of Numeric-Boolean Columns = ", len(num_bool_vars))
        print("    Number of Discrete String Columns = ", len(discrete_string_vars))
        print("    Number of NLP String Columns = ", len(nlp_vars))
        print("    Number of Date Time Columns = ", len(date_vars))
        print("    Number of ID Columns = ", len(id_vars))
        print("    Number of Columns to Delete = ", len(cols_delete))
    if verbose == 2:
        marthas_columns(df_preds,verbose=1)
    if verbose >=1 and orig_cols_total > max_cols_to_print:
        print("    Numeric Columns: %s" %continuous_vars[:max_cols_to_print])
        print("    Integer-Categorical Columns: %s" %int_vars[:max_cols_to_print])
        print("    String-Categorical Columns: %s" %cat_vars[:max_cols_to_print])
        print("    Factor-Categorical Columns: %s" %factor_vars[:max_cols_to_print])
        print("    String-Boolean Columns: %s" %string_bool_vars[:max_cols_to_print])
        print("    Numeric-Boolean Columns: %s" %num_bool_vars[:max_cols_to_print])
        print("    Discrete String Columns: %s" %discrete_string_vars[:max_cols_to_print])
        print("    NLP text Columns: %s" %nlp_vars[:max_cols_to_print])
        print("    Date Time Columns: %s" %date_vars[:max_cols_to_print])
        print("    ID Columns: %s" %id_vars[:max_cols_to_print])
        print("    Columns that will not be considered in modeling: %s" %cols_delete[:max_cols_to_print])
    ##### now collect all the column types and column names into a single dictionary to return!
    #### Since cols_delete and id_vars have the same columns, you need to subtract id_vars from this!
    len_sum_all_cols = reduce(add,[len(v) for v in sum_all_cols.values()]) - len(id_vars)
    if len_sum_all_cols == orig_cols_total:
        print('    %d Predictors classified...' %orig_cols_total)
        #print('        This does not include the Target column(s)')
    else:
        print('Number columns classified %d does not match %d total cols. Continuing...' %(
                   len_sum_all_cols, orig_cols_total))
        ls = sum_all_cols.values()
        flat_list = [item for sublist in ls for item in sublist]
        if len(left_subtract(list(train),flat_list)) == 0:
            print(' Missing columns = None')
        else:
            print(' Missing columns = %s' %left_subtract(list(train),flat_list))
    return sum_all_cols
#################################################################################
import re
WORD = re.compile(r'\w+')
def tokenize_fast(text):
    """
    This is a fast function that tokenizes a string of text into its words
    """
    words = WORD.findall(text)
    return words
############################################################################################
def check_model_options(model_options, name, default):
    try:
        if model_options[name]:
            value = model_options[name]
        else:
            value = default
    except:
        value = default
    return value
#####################################################################################
def classify_features_using_pandas(data_sample, target, model_options={}, verbose=0):
    """
    If you send in a small pandas dataframe with the name of target variable(s), you will get back
    all the features classified by type such as dates, cats, ints, floats and nlps. This is all done using pandas.
    """
    data_sample = copy.deepcopy(data_sample)
    ######   This is where you get the cat_vocab_dict is created in the form of feats_max_min #####
    feats_max_min = nested_dictionary()
    print_features = False
    nlps = []
    bools = []
    ### if a variable has more than this many chars, it will be treated like a NLP variable
    nlp_char_limit = check_model_options(model_options, "nlp_char_limit", 50)
    ### if a variable has more than this limit, it will not be treated like a cat variable #
    cat_limit = check_model_options(model_options, "variable_cat_limit", 30)
    ### Classify features using the previously define function #############
    var_df1 = classify_features(data_sample, target, model_options, verbose=verbose)
    #####  This might be useful for users to know whether to use feature-crosses or not ###
    stri, numi, cat_feature_cross_flag = fast_classify_features(data_sample)
    convert_cols = []
    if len(numi['veryhighcats']) > 0:
        convert_cols =  numi['veryhighcats']
    if convert_cols:
        var_df1['int_vars'] = left_subtract(var_df1['int_vars'], convert_cols)
        var_df1['continuous_vars'] = var_df1['continuous_vars'] + convert_cols
    ########################## Set the default variable types here #############
    dates = var_df1['date_vars']
    cats = var_df1['categorical_vars']
    discrete_strings = var_df1['discrete_string_vars']
    lats = var_df1['lat_vars']
    lons = var_df1['lon_vars']
    ignore_variables = copy.deepcopy(var_df1['cols_delete'])
    all_ints = var_df1['int_vars']
    if isinstance(target, list):
        preds = [x for x in  list(data_sample) if x not in target+ignore_variables]
    else:
        preds = [x for x in  list(data_sample) if x not in [target]+ignore_variables]
    feats_max_min['predictors_in_train'] = copy.deepcopy(preds)
    #### Take(1) always displays only one batch only if num_epochs is set to 1 or a number. Otherwise No print! ########
    #### If you execute the below code without take, then it will go into an infinite loop if num_epochs was set to None.
    if verbose >= 1 and target:
            print(f"printing first five values of {target}: {data_sample[target].values[:5]}")
    if len(preds) <= 30:
        print_features = True
    if print_features and verbose > 1:
        print("printing features and their max, min, datatypes in one batch ")
    ###### Now we do the creation of cat_vocab_dict though it is called feats_max_min here #####
    floats = []
    preds_copy = copy.deepcopy(preds)
    for key in preds_copy:
        if str(data_sample[key].dtype) in ['object', 'category']:
            if type('str') in data_sample[key].map(type).value_counts().index:
                feats_max_min[key]["dtype"] = "string"
            elif data_sample[key].map(type).value_counts().index[0] == int:
                data_sample[key] = data_sample[key].astype(np.int32).values
                feats_max_min[key]["dtype"] = np.int32
                all_ints.append(key)
                if key in cats:
                    cats.remove(key)
                    var_df1['categorical_vars'] = copy.deepcopy(cats)
                elif key in discrete_strings:
                    discrete_strings.remove(key)
                    var_df1['discrete_string_vars'] = copy.deepcopy(discrete_strings)
            elif data_sample[key].map(type).value_counts().index[0] == float:
                data_sample[key] = data_sample[key].astype(np.float32).values
                feats_max_min[key]["dtype"] = np.float32
                floats.append(key)
                if key in cats:
                    cats.remove(key)
                    var_df1['categorical_vars'] = copy.deepcopy(cats)
                elif key in discrete_strings:
                    discrete_strings.remove(key)
                    var_df1['discrete_string_vars'] = copy.deepcopy(discrete_strings)
            elif data_sample[key].map(type).value_counts().index[0] == bool:
                data_sample[key] = data_sample[key].astype(bool).values
                feats_max_min[key]["dtype"] = "bool"
                bools.append(key)
                if key in cats:
                    cats.remove(key)
                    var_df1['categorical_vars'] = copy.deepcopy(cats)
                elif key in discrete_strings:
                    discrete_strings.remove(key)
                    var_df1['discrete_string_vars'] = copy.deepcopy(discrete_strings)
        #### This is not a mistake - you have to test it again. That way we make sure type is safe
        if str(data_sample[key].dtype) in ['object', 'category']:
            if data_sample[key].map(type).value_counts().index[0] == object or data_sample[key].map(type).value_counts().index[0] == str:
                feats_max_min[key]["dtype"] = "string"
        elif data_sample[key].dtype in ['bool']:
            feats_max_min[key]["dtype"] = "bool"
            bools.append(key)
            if key in cats:
                cats.remove(key)
        elif str(data_sample[key].dtype).split("[")[0] in ['datetime64','datetime32','datetime16']:
            feats_max_min[key]["dtype"] = "string"
        elif data_sample[key].dtype in [np.int16, np.int32, np.int64]:
            if key in convert_cols:
                feats_max_min[key]["dtype"] = np.float32
                floats.append(key)
            else:
                feats_max_min[key]["dtype"] = np.int32
        else:
            floats.append(key)
            feats_max_min[key]["dtype"] = np.float32
        if feats_max_min[key]['dtype'] in [np.int16, np.int32, np.int64,
                                np.float16, np.float32, np.float64]:
            ##### This is for integer and float variables #######
            if key in lats+lons:
                ### For lats and lons you need the vocab to create bins using pd.qcut ####
                vocab = data_sample[key].unique()
                feats_max_min[key]["vocab"] = vocab
                feats_max_min[key]['size_of_vocab'] = len(vocab)
                feats_max_min[key]["max"] = max(data_sample[key].values)
                feats_max_min[key]["min"] = min(data_sample[key].values)
            else:
                if feats_max_min[key]['dtype'] in [np.int16, np.int32, np.int64]:
                    vocab = data_sample[key].unique()
                    feats_max_min[key]["vocab"] = vocab
                    feats_max_min[key]['size_of_vocab'] = len(vocab)
                else:
                    ### For the rest of the numeric variables, you just need mean and variance ###
                    vocab = data_sample[key].unique()
                    feats_max_min[key]["vocab_min_var"] = [data_sample[key].mean(), data_sample[key].var()]
                    feats_max_min[key]["vocab"] = vocab
                    feats_max_min[key]['size_of_vocab'] = len(vocab)
            feats_max_min[key]["max"] = max(data_sample[key].values)
            feats_max_min[key]["min"] = min(data_sample[key].values)
        elif feats_max_min[key]['dtype'] in ['bool']:
            ### we are going to convert boolean to float type #####
            vocab = data_sample[key].unique()
            full_array = data_sample[key].values
            full_array = np.array([0.0 if type(x) == float else float(x) for x in full_array])
            ### Don't change the next line even though it appears wrong. I have tested and it works!
            vocab = [0.0 if type(x) == float else float(x) for x in vocab]
            feats_max_min[key]["vocab_min_var"] = [full_array.mean(), full_array.var()]
            feats_max_min[key]["vocab"] = vocab
            feats_max_min[key]['size_of_vocab'] = len(vocab)
        elif feats_max_min[key]['dtype'] in ['string']:
            data_sample[[key]] = data_sample[[key]].fillna("missing")
            data_types = len(data_sample[key].map(type).value_counts())
            if data_types > 1:
                print('\nDATA CLEANING ALERT: Dropping %s since it has %s mixed data types.' %(key, data_types))
                print('    Convert this variable to a single data type and re-run deep_autoviml.')
                ignore_variables.append(key)
                preds.remove(key)
                feats_max_min['predictors_in_train'] = preds
                var_df1['cols_delete'] = copy.deepcopy(ignore_variables)
                if key in cats:
                    cats.remove(key)
                    var_df1['categorical_vars'] = copy.deepcopy(cats)
                elif key in discrete_strings:
                    discrete_strings.remove(key)
                    var_df1['discrete_string_vars'] = copy.deepcopy(discrete_strings)
            if not key in ignore_variables:
                if np.max(data_sample[key].map(len)) >= nlp_char_limit:
                    ### This is for NLP variables. You want to remove duplicates #####
                    if key in dates:
                        continue
                    elif key in cats:
                        cats.remove(key)
                        var_df1['categorical_vars'] = cats
                    elif key in discrete_strings:
                        discrete_strings.remove(key)
                        var_df1['discrete_string_vars'] = discrete_strings
                    print('%s is detected as an NLP variable' %key)
                    if key not in var_df1['nlp_vars']:
                        var_df1['nlp_vars'].append(key)
                    #### Now let's calculate some statistics on this NLP variable ###
                    num_rows_in_data = model_options['DS_LEN']
                    if data_sample.shape[0]*data_sample[key].map(len).mean()/1e6 > 1 or data_sample[key].memory_usage(deep=True)/1e6 > 100:
                        ## number of unique words in a document may be only 10% of the total num of words
                        ### If this number exceeds 1, it means there are 1 million words in that document
                        ### Immediately cap the vocab size to 300,000 - don't measure its vocab!!
                        data_sample = data_sample.sample(frac=0.1, random_state=0)
                    try:
                        vocab = np.concatenate(data_sample[key].map(tokenize_fast))
                    except:
                        vocab = np.concatenate(data_sample[key].map(tokenize_fast).values)
                    vocab = np.unique(vocab).tolist()
                    feats_max_min[key]["vocab"] = vocab
                    try:
                        feats_max_min[key]['seq_length'] = int(data_sample[key].map(len).max())
                        num_words_in_each_row = data_sample[key].map(lambda x: len(x.split(" "))).mean()
                        feats_max_min[key]['size_of_vocab'] = int(num_rows_in_data*num_words_in_each_row)
                    except:
                        feats_max_min[key]['seq_length'] = len(vocab) // num_rows_in_data
                        feats_max_min[key]['size_of_vocab'] = len(vocab)
                else:
                    ### This is for string variables ########
                    ####  Now we select features if they are present in the data set ###
                    num_rows_in_data = model_options['DS_LEN']
                    data_sample[[key]] = data_sample[[key]].fillna("missing")
                    vocab = data_sample[key].unique().tolist()
                    vocab = np.unique(vocab).tolist()
                    #vocab = ['missing' if type(x) != str  else x for x in vocab]
                    feats_max_min[key]["vocab"] = vocab
                    feats_max_min[key]['size_of_vocab'] = len(vocab)
                    feats_max_min[key]['seq_length'] = len(vocab) // num_rows_in_data
        else:
            ####  Now we treat bool and other variable types ###
            #feats_max_min[key]["vocab"] = data_sample[key].unique()
            vocab = data_sample[key].unique()
            #### just leave this as is - it works for other data types
            vocab = ['missing' if type(x) == str  else x for x in vocab]
            feats_max_min[key]["vocab"] = vocab
            feats_max_min[key]['size_of_vocab'] = len(vocab)
            #feats_max_min[key]['size_of_vocab'] = len(feats_max_min[key]["vocab"])
        if print_features and verbose > 1:
            print("  {!r:20s}: {}".format('    sample words from vocab', key, data_sample[key].values[:4]))
            print("  {!r:25s}: {}".format('    size of vocab', feats_max_min[key]["size_of_vocab"]))
            print("  {!r:25s}: {}".format('    max', feats_max_min[key]["max"]))
            print("  {!r:25s}: {}".format('    min', feats_max_min[key]["min"]))
            print("  {!r:25s}: {}".format('    dtype', feats_max_min[key]["dtype"]))
    if not print_features:
        print('Number of variables in dataset is too numerous to print...skipping print')
    ##### Save the variable changes back to the variable type dictionary ##
    var_df1['discrete_string_vars'] = copy.deepcopy(discrete_strings)
    var_df1['categorical_vars'] = copy.deepcopy(cats)
    var_df1['cols_delete'] = ignore_variables

    ##### Make some changes to integer variables to select those with less than certain category limits ##
    ints = [ x for x in all_ints if feats_max_min[x]['size_of_vocab'] > cat_limit and x not in floats]

    int_bools = [ x for x in all_ints if feats_max_min[x]['size_of_vocab'] == 2 and x not in floats]

    int_cats = [ x for x in all_ints if feats_max_min[x]['size_of_vocab'] <= cat_limit and x not in floats+int_bools]

    var_df1['int_vars'] = ints
    var_df1['int_cats'] = int_cats
    var_df1['int_bools'] = int_bools
    var_df1["continuous_vars"] = floats
    var_df1['bools'] = bools
    #### It is better to have a baseline number for the size of the dataset here ########
    feats_max_min['DS_LEN'] = len(data_sample)
    feats_max_min["predictors_in_train"] = preds
    print('    after data cleaning, number of predictors used in modeling = %s' %len(preds))
    ### check if cat_vocab_dict has cat_feature_cross_flag in it ###
    if "cat_feature_cross_flag" in model_options.keys():
        ### If they mistakenly type it as cat_feature_cross_flag then use it ####
        if model_options["cat_feature_cross_flag"]:
            cat_feature_cross_flag = model_options["cat_feature_cross_flag"]
            print('performing feature crossing for %s variables' %cat_feature_cross_flag)
        else:
            model_options["cat_feat_cross_flag"] = cat_feature_cross_flag
            print('Not performing feature crossing for categorical nor integer variables' )
    elif "cat_feat_cross_flag" in model_options.keys():
        if model_options["cat_feat_cross_flag"]:
            cat_feature_cross_flag = model_options["cat_feat_cross_flag"]
            print('performing feature crossing for %s variables' %cat_feature_cross_flag)
        else:
            model_options["cat_feat_cross_flag"] = cat_feature_cross_flag
            print('Not performing feature crossing for categorical nor integer variables' )
    else:
        ### If there is no input for cat_feature_cross_flag, then don't do it ###
        cat_feature_cross_flag = ""
        print('Not performing feature crossing for categorical nor integer variables' )
    return data_sample, var_df1, feats_max_min
############################################################################################
def EDA_classify_and_return_cols_by_type(df1, nlp_char_limit=50):
    """
    EDA stands for Exploratory data analysis. This function performs EDA - hence the name
    ########################################################################################
    This handy function classifies your columns into different types : make sure you send only predictors.
    Beware sending target column into the dataframe. You don't want to start modifying it.
    #####################################################################################
    It returns a list of categorical columns, integer cols and float columns in that order.
    """
    ### Let's find all the categorical excluding integer columns in dataset: unfortunately not all integers are categorical!
    catcols = df1.select_dtypes(include='object').columns.tolist() + df1.select_dtypes(include='category').columns.tolist()
    cats = copy.deepcopy(catcols)
    nlpcols = []
    for each_cat in cats:
        try:
            df1[[each_cat]] = df1[[each_cat]].fillna('missing')
            if df1[each_cat].map(len).max() >=nlp_char_limit:
                nlpcols.append(each_cat)
                catcols.remove(each_cat)
        except:
            continue
    intcols = df1.select_dtypes(include='integer').columns.tolist()
    int_cats = [ x for x in intcols if df1[x].nunique() <= 30 and x not in idcols]
    intcols = left_subtract(intcols, int_cats)
    # let's find all the float numeric columns in data
    floatcols = df1.select_dtypes(include='float').columns.tolist()
    return catcols, int_cats, intcols, floatcols, nlpcols
############################################################################################
def EDA_classify_features(train, target, idcols, nlp_char_limit=50):
    ### Test Labeler is a very important dictionary that will help transform test data same as train ####
    test_labeler = defaultdict(list)

    #### all columns are features except the target column and the folds column ###
    if isinstance(target, str):
        features = [x for x in list(train) if x not in [target]+idcols]
    else:
        ### in this case target is a list and hence can be added to idcols
        features = [x for x in list(train) if x not in target+idcols]

    ### first find all the types of columns in your data set ####
    cats, int_cats, ints, floats, nlps = EDA_classify_and_return_cols_by_type(train[features],
                                                                            nlp_char_limit)

    numeric_features = ints + floats
    categoricals_features = copy.deepcopy(cats)
    nlp_features = copy.deepcopy(nlps)

    test_labeler['categoricals_features'] = categoricals_features
    test_labeler['numeric_features'] = numeric_features
    test_labeler['nlp_features'] = nlp_features

    return cats, int_cats, ints, floats, nlps
#############################################################################################
def left_subtract(l1,l2):
    lst = []
    for i in l1:
        if i not in l2:
            lst.append(i)
    return lst
#############################################################################################
def find_number_bins(series):
    """
    Input must be a pandas series. Otherwise it will blow up. Be careful!
    Returns the recommended number of bins for any Series in pandas
    Input must be a float or integer column. Don't send in alphabetical series!
    """
    return int(np.log2(series.nunique())+1)
#########################################################################################
import re
def find_words_in_list(words, in_list):
    result = []
    for each_word in words:
        for in_src in in_list:
            if re.findall(each_word, in_src):
                result.append(in_src)
    return list(set(result))

#############################################################################################
from collections import defaultdict
def find_latitude_longitude_columns_in_df(df, verbose=0):
    matched_pairs = []
    lats, lat_keywords = find_latitude_columns(df, verbose)
    lons, lon_keywords = find_longitude_columns(df, verbose)
    if len(lats) > 0 and len(lons) > 0:
        if len(lats) > 1:
            for each_lat in lats:
                for each_lon in lons:
                    if lat_keywords[each_lat] == lon_keywords[each_lon]:
                        matched_pairs.append((each_lat, each_lon))
        elif len(lats) == 1 and len(lons) == 1:
            each_lat = lats[0]
            each_lon = lons[0]
            matched_pairs.append((each_lat, each_lon))
    if len(matched_pairs) > 0 and verbose:
        print('Matched pairs of latitudes and longitudes: %s' %matched_pairs)
    return lats, lons, matched_pairs

def find_latitude_columns(df, verbose=0):
    columns = df.select_dtypes(include='float').columns.tolist() + df.select_dtypes(include='object').columns.tolist()
    lat_words = find_words_in_list(['Lat','lat','LAT','Latitude','latitude','LATITUDE'], columns)
    sel_columns = lat_words[:]
    lat_keywords = defaultdict(str)
    if len(columns) > 0:
        for lat_word in columns:
            lati_keyword = find_latitude_keyword(lat_word, columns, sel_columns)
            if not lati_keyword == '':
                if lati_keyword == lat_word:
                    lat_keywords[lat_word] = lati_keyword
                else:
                    lat_keywords[lat_word] = lat_word.replace(lati_keyword,'')
    ###### This is where we find whether they are truly latitudes ############
    print('    possible latitude columns in dataset: %s' %sel_columns)
    sel_columns_copy = copy.deepcopy(sel_columns)
    for sel_col in sel_columns_copy:
        if not lat_keywords[sel_col]:
            sel_columns.remove(sel_col)
     #### If there are any more columns left, then do further analysis #######
    if len(sel_columns) > 0:
        sel_cols_float = df[sel_columns].select_dtypes(include='float').columns.tolist()
        if len(sel_cols_float) > 0:
            for sel_column in sel_cols_float:
                if df[sel_column].isnull().sum() > 0:
                    print('Null values in possible latitude column %s. Removing it' %sel_column)
                    sel_columns.remove(sel_column)
                    continue
                if df[sel_column].min() >= -90 and df[sel_column].max() <= 90:
                    if verbose:
                        print('        %s found as latitude column' %sel_column)
                    if sel_column not in sel_columns:
                        sel_columns.append(sel_column)
                        lati_keyword = find_latitude_keyword(sel_column, columns, sel_columns)
                        if not lati_keyword == '':
                            lat_keywords[lat_word] = sel_column.replace(lati_keyword,'')
                else:
                    sel_columns.remove(sel_column)
        sel_cols_string = df[sel_columns].select_dtypes(include='object').columns.tolist()
        if len(sel_cols_string) > 0:
            for sel_column in sel_cols_string:
                if len(df[df[sel_column].str.endswith(('N','S'))]) > 0:
                    if verbose:
                        print('        %s found as latitude column' %sel_column)
                    if sel_column not in sel_columns:
                        sel_columns.append(sel_column)
                        lati_keyword = find_latitude_keyword(sel_column, columns, sel_columns)
                        if not lati_keyword == '':
                            lat_keywords[lat_word] = sel_column.replace(lati_keyword,'')
                else:
                    sel_columns.remove(sel_column)
    ### after everything what is left is shown here ####
    if len(sel_columns) == 0:
        print('        after further analysis, no latitude columns found')
    else:
        print('        after further analysis, selected latitude columns = %s' %sel_columns)
    return sel_columns, lat_keywords

def find_latitude_keyword(lat_word, columns, sel_columns=[]):
    lat_keywords = defaultdict(str)
    ####  This is where we find the text that is present in column related to latitude ##
    if len(columns) > 0:
        if lat_word.lower() == 'lat':
            if lat_word not in sel_columns:
                sel_columns.append(lat_word)
            lat_keywords[lat_word] = 'lat'
        elif lat_word.lower() == 'latitude':
            if lat_word not in sel_columns:
                sel_columns.append(lat_word)
            lat_keywords[lat_word] = 'latitude'
        elif 'latitude' in lat_word.lower().split(" "):
            if lat_word not in sel_columns:
                sel_columns.append(lat_word)
            lat_keywords[lat_word] = 'latitude'
        elif 'latitude' in lat_word.lower().split("_"):
            if lat_word not in sel_columns:
                sel_columns.append(lat_word)
            lat_keywords[lat_word] = 'latitude'
        elif 'latitude' in lat_word.lower().split("-"):
            if lat_word not in sel_columns:
                sel_columns.append(lat_word)
            lat_keywords[lat_word] = 'latitude'
        elif 'latitude' in lat_word.lower().split("/"):
            if lat_word not in sel_columns:
                sel_columns.append(lat_word)
            lat_keywords[lat_word] = 'latitude'
        elif 'latitude' in lat_word.lower():
            if lat_word not in sel_columns:
                sel_columns.append(lat_word)
            lat_keywords[lat_word] = 'latitude'
        elif 'lat' in lat_word.lower().split(" "):
            if lat_word not in sel_columns:
                sel_columns.append(lat_word)
            lat_keywords[lat_word] = 'lat'
        elif 'lat' in lat_word.lower().split("_"):
            if lat_word not in sel_columns:
                sel_columns.append(lat_word)
            lat_keywords[lat_word] = 'lat'
        elif 'lat' in lat_word.lower().split("-"):
            if lat_word not in sel_columns:
                sel_columns.append(lat_word)
            lat_keywords[lat_word] = 'lat'
        elif 'lat' in lat_word.lower().split("/"):
            if lat_word not in sel_columns:
                sel_columns.append(lat_word)
            lat_keywords[lat_word] = 'lat'
        elif 'lat' in lat_word.lower():
            if lat_word not in sel_columns:
                sel_columns.append(lat_word)
            lat_keywords[lat_word] = 'lat'
    return lat_keywords[lat_word]

def find_longitude_keyword(lon_word, columns, sel_columns=[]):
    lon_keywords = defaultdict(str)
    ####  This is where we find the text that is present in column related to longitude ##
    if len(columns) > 0:
        if lon_word.lower() == 'lon':
            if lon_word not in sel_columns:
                sel_columns.append(lon_word)
            lon_keywords[lon_word] = 'lon'
        elif lon_word.lower() == 'longitude':
            if lon_word not in sel_columns:
                sel_columns.append(lon_word)
            lon_keywords[lon_word] = 'longitude'
        elif 'longitude' in lon_word.lower().split(" "):
            if lon_word not in sel_columns:
                sel_columns.append(lon_word)
            lon_keywords[lon_word] = 'longitude'
        elif 'longitude' in lon_word.lower().split("_"):
            if lon_word not in sel_columns:
                sel_columns.append(lon_word)
            lon_keywords[lon_word] = 'longitude'
        elif 'longitude' in lon_word.lower().split("-"):
            if lon_word not in sel_columns:
                sel_columns.append(lon_word)
            lon_keywords[lon_word] = 'longitude'
        elif 'longitude' in lon_word.lower().split("/"):
            if lon_word not in sel_columns:
                sel_columns.append(lon_word)
            lon_keywords[lon_word] = 'longitude'
        elif 'longitude' in lon_word.lower():
            if lon_word not in sel_columns:
                sel_columns.append(lon_word)
            lon_keywords[lon_word] = 'longitude'
        elif 'lon' in lon_word.lower().split(" "):
            if lon_word not in sel_columns:
                sel_columns.append(lon_word)
            lon_keywords[lon_word] = 'lon'
        elif 'lon' in lon_word.lower().split("_"):
            if lon_word not in sel_columns:
                sel_columns.append(lon_word)
            lon_keywords[lon_word] = 'lon'
        elif 'lon' in lon_word.lower().split("-"):
            if lon_word not in sel_columns:
                sel_columns.append(lon_word)
            lon_keywords[lon_word] = 'lon'
        elif 'lon' in lon_word.lower().split("/"):
            if lon_word not in sel_columns:
                sel_columns.append(lon_word)
            lon_keywords[lon_word] = 'lon'
        elif 'lon' in lon_word.lower():
            if lon_word not in sel_columns:
                sel_columns.append(lon_word)
            lon_keywords[lon_word] = 'lon'
    return lon_keywords[lon_word]

def find_longitude_columns(df, verbose=0):
    columns = df.select_dtypes(include='float').columns.tolist() + df.select_dtypes(include='object').columns.tolist()
    lon_words = find_words_in_list(['Lon','lon','LON','Longitude','Longitude', "LONGITUDE"], columns)
    sel_columns = lon_words[:]
    lon_keywords = defaultdict(str)
    if len(columns) > 0:
        for lon_word in columns:
            long_keyword = find_longitude_keyword(lon_word, columns, sel_columns)
            if not long_keyword == '':
                if lon_word == long_keyword:
                    lon_keywords[lon_word] = long_keyword
                else:
                    lon_keywords[lon_word] = lon_word.replace(long_keyword,'')
    #####  This is where we test whether they are indeed longitude columns ####
    print('    possible longitude columns in dataset: %s' %sel_columns)
    sel_columns_copy = copy.deepcopy(sel_columns)
    for sel_col in sel_columns_copy:
        if not lon_keywords[sel_col]:
            sel_columns.remove(sel_col)
    ###### This is where we find whether they are truly longitudes ############
    if len(sel_columns) > 0:
        sel_cols_float = df[sel_columns].select_dtypes(include='float').columns.tolist()
        if len(sel_cols_float) > 0:
            for sel_column in sel_cols_float:
                if df[sel_column].isnull().sum() > 0:
                    print('Null values in possible longitude column %s. Removing it' %sel_column)
                    sel_columns.remove(sel_column)
                    continue
                if df[sel_column].min() >= -180 and df[sel_column].max() <= 180:
                    if verbose:
                        print('        %s found as longitude column' %sel_column)
                    if sel_column not in sel_columns:
                        sel_columns.append(sel_column)
                        long_keyword = find_longitude_keyword(sel_column, columns, sel_columns)
                        if not long_keyword == '':
                            lon_keywords[lon_word] = sel_column.replace(long_keyword,'')
                else:
                    sel_columns.remove(sel_column)
        sel_cols_string = df[sel_columns].select_dtypes(include='object').columns.tolist()
        if len(sel_cols_string) > 0:
            for sel_column in sel_cols_string:
                if len(df[df[sel_column].str.endswith(('N','S'))]) > 0:
                    if verbose:
                        print('        %s found as longitude column' %sel_column)
                    if sel_column not in sel_columns:
                        sel_columns.append(sel_column)
                        long_keyword = find_longitude_keyword(sel_column, columns, sel_columns)
                        if not long_keyword == '':
                            lon_keywords[lon_word] = sel_column.replace(long_keyword,'')
                else:
                    sel_columns.remove(sel_column)
    ### this is where we finally can select columns ##
    if len(sel_columns) == 0:
        print('        after further analysis, no longitude columns found')
    else:
        print('        after further analysis, selected longitude columns = %s' %sel_columns)
    return sel_columns, lon_keywords
###########################################################################################
from collections import defaultdict
def nested_dictionary():
    return defaultdict(nested_dictionary)
############################################################################################
def classify_dtypes_using_TF2(data_sample, preds, idcols, verbose=0):
    """
    #### This works only on train data sets since they have both features and labels. ################
    ####  It also works in only certain types of tf.data.datasets since every TF dataset is different format.
    If you send in a batch of Ttf.data.dataset with the name of target variable(s), you will get back
    all the features classified by type such as cats, ints, floats and nlps. This is all done using TF2.
    """
    print_features = False
    nlps = []
    nlp_char_limit = 50
    all_ints = []
    floats = []
    cats = []
    int_vocab = 0
    feats_max_min = nested_dictionary()
    #### Take(1) always displays only one batch only if num_epochs is set to 1 or a number. Otherwise No print! ########
    #### If you execute the below code without take, then it will go into an infinite loop if num_epochs was set to None.
    if data_sample.element_spec[0][preds[0]].shape[0] is None:
        for feature_batch, label_batch in data_sample.take(1):
            if verbose >= 1:
                print(f"{target}: {label_batch[:4]}")
            if len(feature_batch.keys()) <= 30:
                print_features = True
                if verbose >= 1:
                    print("features and their max, min, datatypes in one batch of size: ",batch_size)
                for key, value in feature_batch.items():
                    feats_max_min[key]["dtype"] = data_sample.element_spec[0][key].dtype
                    if feats_max_min[key]['dtype'] in [tf.float16, tf.float32, tf.float64]:
                        ## no need to find vocab of floating point variables!
                        floats.append(key)
                    elif feats_max_min[key]['dtype'] in [tf.int16, tf.int32, tf.int64]:
                        ### if it is an integer var, it is worth finding their vocab!
                        all_ints.append(key)
                        int_vocab = tf.unique(value)[0].numpy().tolist()
                        feats_max_min[key]['size_of_vocab'] = len(int_vocab)
                    elif feats_max_min[key]['dtype'] in [tf.string]:
                        feature_batch[[key]] = feature_batch[[key]].fillna("missing")
                        if tf.reduce_max(tf.strings.length(feature_batch[key])).numpy() >= nlp_char_limit:
                            print('%s is detected and will be treated as an NLP variable')
                            nlps.append(key)
                        else:
                            cats.append(key)
            if not print_features:
                print('Number of variables in dataset is too numerous to print...skipping print')

    ints = [ x for x in all_ints if feats_max_min[x]['size_of_vocab'] > 30 and x not in idcols]

    int_cats = [ x for x in all_ints if feats_max_min[x]['size_of_vocab'] <= 30 and x not in idcols]

    return cats, int_cats, ints, floats, nlps
############################################################################################
# Define feature columns(Including feature engineered ones )
# These are the features which come from the TF Data pipeline
def create_feature_cols(data_batches, preds):
    #Keras format features
    keras_dict_input = {}
    if data_batches.element_spec[0][preds[0]].shape[0] is None:
        print("Creating keras features dictionary...")
        for feature_batch, label_batch in data_batches.take(1):
            for key, _ in feature_batch.items():
                k_month = tf.keras.Input(name=key, shape=(1,), dtype=tf.string)
                keras_dict_input[key] = k_month
    print('    completed.')
    return({'K' : keras_dict_input})
##############################################################################################
# Removes duplicates from a list to return unique values - USED ONLYONCE
def find_remove_duplicates(values):
    output = []
    seen = set()
    for value in values:
        if value not in seen:
            output.append(value)
            seen.add(value)
    return output
#################################################################################
from collections import defaultdict
import copy
def fast_classify_features(df):
    """
    This is a very fast way to get a handle on what a dataset looks like. Just send in df and get a print.
    Nothing is returned. You just get a printed number of how many types of features you have in dataframe.
    """
    num_list = df.select_dtypes(include='integer').columns.tolist()
    float_list = df.select_dtypes(include='float').columns.tolist()
    str_list = left_subtract(df.columns.tolist(), num_list+float_list)
    all_list = [str_list, num_list]
    str_dict = defaultdict(dict)
    int_dict = defaultdict(dict)
    for inum, dicti in enumerate([str_dict, int_dict]):
        bincols = []
        catcols = []
        highcols = []
        numcols = []
        for col in all_list[inum]:
            leng = len(df[col].value_counts())
            if leng <= 2:
                bincols.append(col)
            elif leng > 2 and leng <= 15:
                catcols.append(col)
            elif leng >15 and leng <100:
                highcols.append(col)
            else:
                numcols.append(col)
        dicti['bincols'] = bincols
        dicti['catcols'] = catcols
        dicti['highcats'] = highcols
        dicti['veryhighcats'] = numcols
        if inum == 0:
            str_dict = copy.deepcopy(dicti)
            print('Distribution of string columns in datatset:')
            print('    number of binary = %d, cats = %d, high cats = %d, very high cats = %d' %(
                len(bincols), len(catcols), len(highcols), len(numcols)))
        else:
            print('Distribution of integer columns in datatset:')
            int_dict = copy.deepcopy(dicti)
            print('    number of binary = %d, cats = %d, high cats = %d, very high cats = %d' %(
                len(bincols), len(catcols), len(highcols), len(numcols)))
    ###   Check if worth doing cat_feature_cross_flag on this dataset ###
    int_dict['floats'] = float_list
    print('Distribution of floats:')
    print('    number float variables = %d' %len(float_list))
    cat_feature_cross_flag = []
    if len(str_dict['bincols']+str_dict['catcols']) > 2 and len(str_dict['bincols']+str_dict['catcols']) <= 10:
        cat_feature_cross_flag.append("cat")
    if len(int_dict['bincols']+int_dict['catcols']) > 2 and len(int_dict['bincols']+int_dict['catcols']) <= 10:
        cat_feature_cross_flag.append("num")
    ######## This is where we advise and act on whether to do feature crosses or not #########
    if cat_feature_cross_flag:
        print('Data Transformation Advisory:')
        if "cat" in cat_feature_cross_flag:
            cat_feature_cross_flag = "cat"
            print('    performing categorical feature crosses: changed cat_feat_cross_flag to "cat"')
        elif "num" in cat_feature_cross_flag:
            cat_feature_cross_flag = "num"
            print('    performing integer feature crosses: changed cat_feat_cross_flag to "num" ')
        elif "cat" in cat_feature_cross_flag and "num" in cat_feature_cross_flag:
            cat_feature_cross_flag = "both"
            print('    performing both integer and cat feature crosses: changed cat_feat_cross_flag to "both" ')
    else:
        print('No data transformations needed in this dataset')
        cat_feature_cross_flag = ""
    if len(int_dict['veryhighcats']) > 0:
        print('    we transformed %s from integer to float' %int_dict['veryhighcats'])
    return str_dict, int_dict, cat_feature_cross_flag
###################################################################################################



############################################################################################
#Copyright 2021 Google LLC

#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
############################################################################################
import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="deep_autoviml",
    version="0.0.84",
    author="Ram Seshadri",
    # author_email="author@example.com",
    description="Automatically Build Deep Learning Models and Pipelines fast!",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license='Apache License 2.0',
    url="https://github.com/AutoViML/deep_autoviml",
    packages = [
        "deep_autoviml",
        "deep_autoviml.data_load",
        "deep_autoviml.modeling",
        "deep_autoviml.models",
        "deep_autoviml.preprocessing",
        "deep_autoviml.utilities",
    ],
    include_package_data=True,
    install_requires=[
        "ipython",
        "jupyter",
        "tensorflow~=2.5",
        "pandas",
        "matplotlib",
        "numpy~=1.19.2",
        "scikit-learn>=0.23.1, <=0.24.2",
        "regex",
        "emoji",
        "storm-tuner>=0.0.8",
        "optuna",
        "tensorflow_hub~=0.12.0",
        "tensorflow-text~=2.5",
        "xlrd"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)



############################################################################################
#Copyright 2021 Google LLC

#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
############################################################################################
import tensorflow as tf
from tensorflow import keras
#### Make sure it is Tensorflow 2.4 or greater!
from tensorflow.keras.optimizers import SGD, Adam, RMSprop
from tensorflow.keras import layers
from tensorflow.keras import optimizers
from tensorflow.keras import models
from tensorflow.keras import callbacks
from tensorflow.keras import backend as K
from tensorflow.keras import utils
from tensorflow.keras import layers
from tensorflow.keras.layers import BatchNormalization
from tensorflow.keras.optimizers import SGD
from tensorflow.keras import regularizers
from tensorflow.keras.layers import Reshape, MaxPooling1D, MaxPooling2D
from tensorflow.keras.layers import AveragePooling2D, AveragePooling1D
from tensorflow.keras import Model, Sequential
from tensorflow.keras.layers import Embedding, Reshape, Dropout, Dense, SimpleRNN, LeakyReLU
from tensorflow.keras.layers import Activation, Dense, Embedding, GlobalAveragePooling1D
from tensorflow.keras.layers import GlobalMaxPooling1D, Dropout, Conv1D
from tensorflow.keras.layers.experimental.preprocessing import TextVectorization
############################################################################################


def make_rnn(model_options):
    '''
    Author: Adarsh C
    Date created: 30/01/2022
    Date last modified: 30/01/2022
    contact: chekodu.adarsh@gmail.com
    Inputs:
    model_options: contains important model hyper parameters
    '''
    model = tf.keras.Sequential()
    model.add(SimpleRNN(128, input_shape= (model_options['window_length'], len(model_options['features'])), return_sequences=True))
    model.add(LeakyReLU(alpha=0.5))
    model.add(SimpleRNN(128, return_sequences=True))
    model.add(LeakyReLU(alpha=0.5)) 
    model.add(Dropout(0.3)) 
    model.add(SimpleRNN(64, return_sequences=False))
    model.add(Dropout(0.3)) 
    model.add(Dense(1))
    return model